import numpy as np
import scipy
import tinyarray as ta
import scipy.linalg as la
from itertools import product
from copy import copy, deepcopy
from numbers import Number
from warnings import warn
from functools import lru_cache
import sympy
from sympy.core.basic import Basic
from sympy.core.function import AppliedUndef
from collections import defaultdict, abc, UserDict
from .linalg import prop_to_id, allclose

from . import kwant_continuum

_commutative_momenta = [kwant_continuum.make_commutative(k, k)
           for k in kwant_continuum.momentum_operators]

e = kwant_continuum.sympify('e')
I = kwant_continuum.sympify('I')


def substitute_exponents(expr):
    """Substitute trignometric functions with exp.

    sin(X) -> (e^(I * X) - e^(-I * X)) / (2 * I)
    cos(X) -> (e^(I * X) + e^(-I * X)) / 2
    exp(X) -> e^X
    """

    subs = {}
    for f in expr.atoms(AppliedUndef, sympy.Function):
        # if more than one argument, we continue
        if len(f.args) > 1:
            continue
        else:
            arg = f.args[0]

        # if only one argument, we follow with subs
        if str(f.func) == 'sin':
            subs[f] = (e**(I * arg) - e**(-I * arg)) / (2 * I)

        if str(f.func) == 'cos':
            subs[f] = (e**(I * arg) + e**(-I * arg)) / 2

        if str(f.func) == 'exp':
            subs[f] = e**arg

    return expr.subs(subs).expand()


class BlochCoeff(tuple):

    def __new__(cls, hop, coeff):
        """
        Container for Bloch coefficient in ``BlochModel``, in the form of
        ``(hop, coeff)``, equivalent to ``coeff * exp(I * hop.dot(k))``.
        """
        if not (isinstance(hop, np.ndarray) and isinstance(coeff, sympy.Expr)):
            raise ValueError('`hop` must be a 1D numpy array and `coeff` a sympy expression.')
        if isinstance(coeff, sympy.add.Add):
            raise ValueError('`coeff` must be a single term with no sum.')
        return super(BlochCoeff, cls).__new__(cls, [hop, coeff])

    def __hash__(self):
        # only hash coeff
        return hash(self[1])

    def __eq__(self, other):
        hop1, coeff1 = self
        if other == 1:
            return allclose(hop1, 0) and coeff1 == 1
        hop2, coeff2 = other
        # test equality of hop with allclose
        return allclose(hop1, hop2) and coeff1 == coeff2

    def __mul__(self, other):
        hop1, coeff1 = self
        if isinstance(other, sympy.Expr):
            return BlochCoeff(hop1, coeff1 * other)
        elif isinstance(other, BlochCoeff):
            hop2, coeff2 = other
            return BlochCoeff(hop1 + hop2, coeff1 * coeff2)
        else:
            raise NotImplementedError

    def __rmul__(self, other):
        hop1, coeff1 = self
        if isinstance(other, sympy.Expr):
            return BlochCoeff(hop1, other * coeff1)
        else:
            raise NotImplementedError

    def __deepcopy__(self, memo):
        hop, coeff = self
        return BlochCoeff(deepcopy(hop), deepcopy(coeff))

    def __copy__(self):
        hop, coeff = self
        return BlochCoeff(copy(hop), copy(coeff))

    def tosympy(self, momenta, nsimplify=False):
        hop, coeff = self
        if nsimplify:
            # Vectorize nsimplify
            vnsimplify = np.vectorize(sympy.nsimplify, otypes=[object])
            hop = vnsimplify(hop)
        return coeff * e**(sum(I * ki * di for ki, di in zip(momenta, hop)))


class Model(UserDict):

    # Make it work with numpy arrays
    __array_ufunc__ = None

    def __init__(
        self,
        hamiltonian=None,
        locals=None,
        momenta=('k_x', 'k_y', 'k_z'),
        keep=None,
        symbol_normalizer=None, normalize=False, shape=None, format=None
    ):
        """
        Symbolic matrix-valued function that depends on momenta and other parameters.

        Implements the algebra of matrix valued functions.
        Implements many sympy and numpy methods and overrides arithmetic operators.
        Internally it represents ``sum(symbol * value)``, where ``symbol`` is a symbolic
        expression, and ``value`` can be scalar, array (both dense and sparse)
        or LinearOperator. This is accessible as a dict ``{symbol: value}``.

        Parameters
        ----------
        hamiltonian : str, SymPy expression, dict or None (default)
            Symbolic representation of a Hamiltonian.  If a string, it is
            first converted to a SymPy expression using `kwant_continuum.sympify`.
            If a dict is provided, it should have the form
            ``{symbol: array}`` with all arrays the same size (dense or sparse).
            ``symbol`` by default is passed through sympy.sympify, and should
            consist purely of a product of symbolic coefficients, no constant
            factors other than 1, except if ``normalize=True``. ``None`` initializes
            a zero ``Model``.
        locals : dict or ``None`` (default)
            Additional namespace entries for `~kwant_continuum.sympify`.  May be
            used to simplify input of matrices or modify input before proceeding
            further. For example:
            ``locals={'k': 'k_x + I * k_y'}`` or
            ``locals={'sigma_plus': [[0, 2], [0, 0]]}``.
        keep : iterable of expressions (optional)
            Set of symbolic coefficients that are kept, anything that does not
            appear here is discarded. Useful for perturbative calculations where
            only terms to a given order are needed. By default all keys are kept.
        momenta : iterable of strings or Sympy symbols
            Names of momentum variables, default ``['k_x', 'k_y', 'k_z']`` or
            corresponding sympy symbols. Momenta are treated the same as other
            keys for the purpose of `keep`.
        symbol_normalizer : callable (optional)
            Function to apply symbols when initializing with dict. By default the
            keys are passed through ``sympy.sympify`` and ``sympy.expand_power_exp``.
            ``keep`` are also passed through ``symbol_normalizer``.
        normalize : bool, default False
            Whether to clean input dict by splitting summands in symbols,
            moving numerical factors in the symbols to values, removing entries
            with values allclose to zero. Ignored if hamiltonian is not a dict.
        shape : tuple or None (default)
            Shape of the Model, must match the shape of all the values. If not
            provided, it is automatically found based on the shape of the input.
            Must be provided if ``hamiltonian`` is ``None`` or ``{}``. Empty tuple
            corresponds to scalar values.
        format : class or None (default)
            Type of the values in the model. Supported types are
            ``np.complex128``, ``scipy.sparse.linalg.LinearOperator``, ``np.ndarray``,
            and subclasses of ``scipy.sparse.spmatrix`` . If ``hamiltonian`` is
            provided as a dict, all values must be of this type, except for
            scalar values, which are recast to ``np.complex128``. If ``format`` is
            not provided, it is inferred from the type of the values. Must be
            provided if ``hamiltonian`` is `None` or ``{}``. If ``hamiltonian`` is
            not a dictionary, ``format`` is ignored an set to ``np.ndarray``.
        """
        if hamiltonian is None:
            hamiltonian = {}
        if symbol_normalizer is None:
            symbol_normalizer = _symbol_normalizer
        self.momenta = _find_momenta(tuple(momenta))

        if keep is not None:
            self.keep = {symbol_normalizer(k) for k in keep}
        else:
            self.keep = set()

        if hamiltonian == {} or isinstance(hamiltonian, abc.Mapping):
            # Initialize as dict sympifying the keys
            self.data = {symbol_normalizer(k): v for k, v in hamiltonian.items()
                              if symbol_normalizer(k) in self.keep
                                 or not self.keep}

        else:
            # Try to parse the input with kwant_continuum.sympify
            hamiltonian = kwant_continuum.sympify(hamiltonian, locals=locals)
            if not isinstance(hamiltonian, sympy.matrices.MatrixBase):
                hamiltonian = sympy.Matrix([[hamiltonian]])
            hamiltonian = substitute_exponents(hamiltonian)
            free_parameters = list(hamiltonian.atoms(sympy.Symbol))
            gens = free_parameters + list(self.momenta)
            hamiltonian = kwant_continuum.make_commutative(hamiltonian, *gens)
            monomials = kwant_continuum.monomials(hamiltonian)
            monomials = {k: kwant_continuum.lambdify(v)()
                         for k, v in monomials.items()}
            # remove matrices == zeros
            monomials = {k: v for k, v in monomials.items()
                         if not np.allclose(v, 0)}
            self.data = monomials
            normalize = True

        # Find shape and format
        self.shape = shape
        self.format = format
        if self.shape is None or self.format is None:
            if self.data == {}:
                # raise ValueError('Must provide `shape` and `format` when initializing empty Model.')
                warn('Provide `shape` and `format` when initializing empty Model.', DeprecationWarning)
            else:
                val = next(iter(self.values()))
                shape, format = _shape_and_format(val)
                self.shape = (shape if self.shape is not None else shape)
                self.format = (format if self.format is not None else format)
        if shape == ():
            # Recast numbers to np.complex128
            self.data = {k: np.complex128(v) for k, v in self.items()}
        if not all(issubclass(type(v), format) for v in self.values()):
            raise ValueError('All values must have the same `format`.')
        if not all(v.shape == shape for v in self.values()):
            raise ValueError('All values must have the same `shape`.')

        if normalize:
            # Clean internal data by:
            # * splitting summands in keys
            # * moving numerical factors to values
            # * removing entries which values care np.allclose to zero
            old_data = {copy(key): copy(val) for key, val in self.items()}
            self.data = {}
            for key, val in old_data.items():
                for summand in key.expand().powsimp(combine='exp').as_ordered_terms():
                    factors = summand.as_ordered_factors()
                    symbols, numbers = [], []
                    for f in factors:
                        # This catches sqrt(2) and much faster than f.is_constant()
                        if f.is_number:
                            numbers.append(f)
                        else:
                            symbols.append(f)
                    new_key = sympy.Mul(*symbols)
                    new_val = complex(sympy.Mul(*numbers))  * val
                    self[new_key] += new_val
            # remove zero entries
            self.data = {k: v for k, v in self.items() if not allclose(v, 0)}

    # Make sure values have the correct format
    def __setitem__(self, key, item):
        if (isinstance(item, self.format) and self.shape == item.shape):
            self.data[key] = item
        elif (isinstance(item, Number) and self.shape == ()):
            self.data[key] = np.complex128(item)
        else:
            raise ValueError('Format of item ({}) must match the format ({}) '
                             'and shape ({}) of Model'.format(item, self.format, self.shape))

    # Allow getting values using text keys
    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        elif _symbol_normalizer(key) in self.data:
            return self.data[_symbol_normalizer(key)]
        else:
            return self.__missing__(key)

    # Defaultdict functionality
    def __missing__(self, key):
        if self.format is np.complex128:
            #scalar
            return np.complex128(0)
        elif self.format is np.ndarray:
            # Return dense zero array if dense
            return np.zeros(self.shape, dtype=complex)
        elif issubclass(self.format, scipy.sparse.spmatrix):
            # Return a zero sparse matrix of the same type
            return self.format(self.shape, dtype=complex)
        elif issubclass(self.format, scipy.sparse.linalg.LinearOperator):
            return scipy.sparse.linalg.aslinearoperator(
                scipy.sparse.csr_matrix(self.shape, dtype=complex))

    def __eq__(self, other):
        # Call allclose with default tolerances
        return self.allclose(other)

    def __add__(self, other):
        # Addition of Models. It is assumed that both Models are
        # structured correctly, every key is in standard form.

        # Useful for sum to work.
        if other is 0:
            result = self.copy()
        # Temporarily allow adding malshaped empty Models
        elif (isinstance(other, type(self)) and other.data=={}):
            result = self.copy()
        elif (isinstance(other, type(self)) and self.data=={}):
            result = other.copy()
        elif isinstance(other, type(self)):
            if not (self.format is other.format and self.shape == other.shape):
                raise ValueError('Addition is only possible for Models with the same shape and data type.')
            # other is not empty, so the result is not empty
            if self.momenta != other.momenta:
                raise ValueError("Can only add Models with the same momenta")
            result = self.zeros_like()
            for key in self.keys() & other.keys():
                result[key] = self[key] + other[key]
            for key in self.keys() - other.keys():
                result[key] = copy(self[key])
            for key in other.keys() - self.keys():
                result[key] = copy(other[key])
        elif ((isinstance(other, self.format) and self.shape == other.shape)
              or (isinstance(other, Number) and self.shape == ())):
            # Addition of constants
            result = self.copy()
            result[1] += other
        else:
            raise NotImplementedError('Addition of {} with shape {} with {} not supported'.format(type(self), self.shape, type(other)))
        return result

    def __radd__(self, other):
        # Addition of monomials with other types.

        # Useful for sum to work.
        if not other:
            result = self.copy()
        elif ((isinstance(other, self.format) and self.shape == other.shape)
              or (isinstance(other, Number) and self.shape == ())):
            # Addition of constants
            result = self.copy()
            result[1] += other
        else:
            raise NotImplementedError('Addition of {} with {} not supported'.format(type(self), type(other)))
        return result

    def __neg__(self):
        result = self.zeros_like()
        result.data = {key: -val for key, val in self.items()}
        return result

    def __sub__(self, other):
        return self + (-other)

    def __mul__(self, other):
        # Multiplication by numbers, sympy symbols, arrays and Model
        if isinstance(other, Number):
            result = self.zeros_like()
            result.data = {key: val * other for key, val in self.items()}
        elif isinstance(other, Basic):
            result = sum((type(self)({key * other: copy(val)},
                                     keep=keep,
                                     momenta=self.momenta)
                         for key, val in self.items()
                         if (key * other in keep or not keep)),
                         self.zeros_like())
        elif isinstance(other, Model):
            if not (issubclass(self.format, (Number, np.ndarray)) or
                    issubclass(other.format, (Number, np.ndarray))):
                raise ValueError('Elementwise multiplication only allowed for scalar '
                                 'and ndarra data types. With sparse matrices use `@` '
                                 'for matrix multiplication.')
            if self.momenta != other.momenta:
                raise ValueError("Can only multiply Models with the same momenta")
            keep = self.keep | other.keep
            result = sum(type(self)({k1 * k2: v1 * v2},
                                    keep=keep,
                                    momenta=self.momenta)
                          for (k1, v1), (k2, v2) in product(self.items(), other.items())
                          if (k1 * k2 in keep or not keep))
            # Find out the shape of the result even if it is empty
            if result is 0:
                result = self.zeros_like()
                result.shape, result.format = _shape_and_format(self[1] * other[1])
        else:
            # Otherwise try to multiply every value with other
            result = self.zeros_like()
            result.data = {key: val * other for key, val in self.items()}
            result.shape, result.format = _shape_and_format(self[1] * other)
        return result

    def __rmul__(self, other):
        # Left multiplication by numbers, sympy symbols and arrays
        if isinstance(other, Number):
            result = self.__mul__(other)
        elif isinstance(other, Basic):
            result = sum((type(self)({other * key: copy(val)},
                                     keep=self.keep,
                                     momenta=self.momenta)
                         for key, val in self.items()
                         if (key * other in self.keep or not self.keep)),
                         self.zeros_like())
        else:
            # Otherwise try to multiply every value with other
            result = self.zeros_like()
            result.data = {key: other * val for key, val in self.items()}
            result.shape, result.format = _shape_and_format(other * self[1])
        return result

    def __matmul__(self, other):
        # Multiplication by arrays and Model
        if isinstance(other, Model):
            if self.momenta != other.momenta:
                raise ValueError("Can only multiply Models with the same momenta")
            keep = self.keep | other.keep
            result = sum(type(self)({k1 * k2: v1 @ v2},
                                    keep=keep,
                                    momenta = self.momenta)
                          for (k1, v1), (k2, v2) in product(self.items(), other.items())
                          if (k1 * k2 in keep or not keep))
            # Find out the shape of the result even if it is empty
            if result is 0:
                result = self.zeros_like()
                result.shape, result.format = _shape_and_format(self[1] @ other[1])
        else:
            # Otherwise try to multiply every value with other
            result = self.zeros_like()
            result.data = {key: val @ other for key, val in self.items()}
            result.shape, result.format = _shape_and_format(self[1] @ other)
        return result

    def __rmatmul__(self, other):
        # Left multiplication by arrays
        result = self.zeros_like()
        result.data = {key: other @ val for key, val in self.items()}
        result.shape, result.format = _shape_and_format(other @ self[1])
        return result

    def __truediv__(self, other):
        result = self.zeros_like()
        if isinstance(other, Number):
            result.data = {key : val * (1/other) for key, val in self.items()}
        else:
            raise TypeError(
                "unsupported operand type for /: {} and {}".format(type(self), type(other)))
        return result

    def __repr__(self):
        result = ['{']
        for k, v in self.items():
            result.extend([str(k), ':\n', str(v), ',\n\n'])
        result.append('}')
        return "".join(result)

    def zeros_like(self):
        """Return an empty model object that inherits the other properties"""
        result = type(self)(shape=self.shape,
                            format=self.format)
        result.keep=self.keep.copy()
        result.momenta=self.momenta
        return result

    def transform_symbolic(self, func):
        """Transform keys by applying func to all of them. Useful for
        symbolic substitutions, differentiation, etc."""
        # Add possible duplicate keys that only differ in constant factors
        result = sum((type(self)({func(key): copy(val)},
                                 normalize=True,
                                 momenta=self.momenta)
                         for key, val in self.items()),
                     self.zeros_like())
        return result

    def rotate_momenta(self, R):
        """Rotate momenta with rotation matrix R."""
        momenta = self.momenta
        assert len(momenta) == R.shape[0], (momenta, R)

        k_prime = R @ sympy.Matrix(momenta)
        rotated_subs = {k: k_prime for k, k_prime in zip(momenta, k_prime)}

        def trf(key):
            return key.subs(rotated_subs, simultaneous=True)

        return self.transform_symbolic(trf)

    def subs(self, *args, **kwargs):
        """Substitute symbolic expressions. See documentation of
        ``sympy.Expr.subs()`` for details.

        Allows for the replacement of momenta in the Model object.
        Replacing a momentum k with a sympy.Symbol object p replaces
        the momentum k with p in the Model.
        Replacing a momentum k with a number removes the momentum k
        from the Model momenta.
        Replacing a momentum k with a sympy expression that does not contain
        any of the Model.momenta also removes the momentum k from the
        momenta.
        """
        # Allowed inputs are an old, new pair, or
        # a list or dictionary of old, new pairs.
        # Bring them all to the form of a list of old, new pairs.
        if len(args) == 2:  # Input is a single pair
            args = ([(args[0], args[1])], )
        elif isinstance(args[0], dict): # Input is a dictionary
            args = ([(key, value) for key, value in args[0].items()], )

        momenta = self.momenta
        for (old, new) in args[0]:
            # Substitution of a momentum variable with a symbol
            # is a renaming of the momentum.
            if old in momenta and isinstance(new, sympy.Symbol):
                momenta = tuple(momentum if old is not momentum else new
                                for momentum in momenta)
            # If no momenta appear in the replacement for a momentum, we consider
            # that momentum removed.
            # Replacement is not a sympy object.
            elif not isinstance(new, sympy.Basic):
                momenta = tuple(momentum for momentum in momenta if old is not momentum)
            # Replacement is a sympy object, but does not contain momenta.
            elif not any([momentum in new.atoms() for momentum in momenta]):
                momenta = tuple(momentum for momentum in momenta if old is not momentum)
        substituted = self.transform_symbolic(lambda x: x.subs(*args, **kwargs))
        substituted.momenta = momenta
        # If there are exponentials, evaluate any numerical exponents,
        # so they can be moved to the matrix valued part of the Model
        result = substituted.zeros_like()
        for key, value in substituted.items():
            # Expand sums in the exponent to products of exponentials,
            # find all exponentials.
            key = sympy.expand(key, power_base=True, power_exp=True,
                               mul=True, log=False, multinomial=False)
            find_expos = [ele for ele in key.args if ele.is_Pow]
            if len(find_expos):
                rest = key / np.prod(find_expos)
                # If an exponential evaluates to a number, replace it with that number.
                # Otherwise, leave the exponential unchanged.
                expos = [expo.subs(e, np.e).evalf() if expo.subs(e, np.e).evalf().is_number
                         else expo for expo in find_expos]
                result += type(substituted)({rest * np.prod(expos): value}, momenta=momenta, normalize=True)
            else:
                result += type(substituted)({key: value}, momenta=momenta, normalize=True)
        return result

    def conj(self):
        """Complex conjugation"""
        result = self.zeros_like()
        # conjugation is bijective, if self was properly formatted, so is this
        result.data = {key.subs(sympy.I, -sympy.I): val.conj()
                        for key, val in self.items()}
        return result

    def T(self):
        """Transpose"""
        result = self.zeros_like()
        result.data = {key: val.T for key, val in self.items()}
        result.shape = self.shape[::-1]
        return result

    def trace(self):
        """Take trace of the matrix and return a scalar valued Model."""
        result = self.zeros_like()
        result.data = {key: np.sum(val.diagonal()) for key, val in self.items()}
        result.shape, result.format = (), np.complex128
        return result

    def value_list(self, key_list):
        """Return a list of the matrix coefficients corresponding to the keys in key_list."""
        return [self[key] for key in key_list]

    def around(self, decimals=3):
        """Return Model with matrices rounded to given number of decimals."""
        result = self.zeros_like()
        for key, val in self.items():
            val = np.around(val, decimals)
            if not np.allclose(val, 0):
                result[key] = val
        return result

    def tosympy(self, nsimplify=False):
        """Return sympy representation of the Model.
        If nsimplify=True, attempt to rewrite numerical coefficients as exact formulas."""
        if not nsimplify:
            result = sympy.sympify(sum(key * val for key, val in self.toarray().items()))
        else:
            # Vectorize nsimplify
            vnsimplify = np.vectorize(sympy.nsimplify, otypes=[object])
            result = sympy.MatAdd(*[key * sympy.Matrix(vnsimplify(val))
                                    for key, val in self.toarray().items()]).doit()
        if isinstance(result, (sympy.MatrixBase,
                               sympy.ImmutableDenseMatrix,
                               sympy.ImmutableDenseNDimArray)):
            result = sympy.Matrix(result).reshape(*result.shape)
        return result

    def evalf(self, subs=None):
        """Evaluate using parameter values in `subs`."""
        return sum(float(key.evalf(subs=subs)) * val for key, val in self.items())

    def tocsr(self):
        """Convert to sparse csr format."""
        result = self.zeros_like()
        result.format = scipy.sparse.csr_matrix
        for key, val in self.items():
            if isinstance(val, (Number, np.ndarray, scipy.sparse.spmatrix)):
                result[key] = scipy.sparse.csr_matrix(val, dtype=complex)
            else:
                # LinearOperator doesn't support multiplication with sparse matrix
                val = scipy.sparse.csr_matrix(val @ np.eye(val.shape[-1], dtype=complex), dtype=complex)
        return result

    def toarray(self):
        """Convert to dense numpy ndarray format."""
        result = self.zeros_like()
        result.format = np.ndarray
        for key, val in self.items():
            if isinstance(val, np.ndarray):
                result[key] = val
            elif isinstance(val, Number):
                result[key] = np.asarray(val)
            elif scipy.sparse.spmatrix:
                result[key] = val.A
            else:
                 val = val @ np.eye(val.shape[-1], dtype=complex)
        return result

    def copy(self):
        """Return a copy."""
        result = self.zeros_like()
        # This is faster than deepcopy of the dict
        result.data = {copy(k): copy(v) for k, v in self.items()}
        return result

    def lambdify(self, nsimplify=False, *, onsite=False, hopping=False):
        """Return a callable object for the model, with sympy symbols as
        parameters.

        Parameters
        ----------
        nsimplify: bool, default False
            Whether or not to attempt to rewrite numerical coefficients as
            exact symbols in sympification.
        onsite : bool, default False
            If True, adds 'site' as the first argument to the callable object.
            Helpful for passing Model objects to kwant Builder objects as
            onsite functions.
        hopping : bool, default False
            If True, adds 'site1' and 'site2' as the first two arguments to
            the callable object.
            Helpful for passing Model objects to kwant Builder objects as
            hopping functions.

        Notes:
        onsite and hopping are mutually exclusive. If both are set to True,
        an error is thrown.
        """
        # Replace 'e' with the numerical value
        expr = self.tosympy(nsimplify=nsimplify).subs({'e': np.e})
        # Needed if expr is an array with 1 element, because .tosympy
        # returns a scalar then.
        try:
            expr = sympy.Matrix(expr).reshape(*expr.shape)
        except TypeError:
            pass
        args = sorted([s.name for s in expr.atoms(sympy.Symbol)])
        if onsite and not hopping:
            args = ['site'] + args
        elif hopping and not onsite:
            args = ['site1', 'site2'] + args
        elif hopping and onsite:
            raise ValueError("'hopping' and 'onsite' are mutually exclusive")
        return sympy.lambdify(args, expr)

    def reshape(self, *args, **kwargs):
        """Reshape, see numpy.reshape."""
        result = self.zeros_like()
        result.data = {key: val.reshape(*args, **kwargs) for key, val in self.items()}
        result.shape, _ = _shape_and_format(self[1].reshape(*args, **kwargs))
        return result

    def allclose(self, other, rtol=1e-05, atol=1e-08, equal_nan=False):
        """Test whether two Models are approximately equal"""
        if other == {} or other == 0:
            if self.data == {}:
                return True
            else:
                return all(allclose(val, 0, rtol, atol, equal_nan) for val in self.values())
        else:
            return all(allclose(self[key], other[key], rtol, atol, equal_nan)
                       for key in self.keys() | other.keys())


class BlochModel(Model):
    def __init__(self, hamiltonian=None, locals=None, momenta=('k_x', 'k_y', 'k_z'),
                 keep=None, shape=None, format=None):
        """
        A ``Model`` where coefficients are periodic functions of momenta.

        Internally it is a ``sum(BlochCoeff * value)``, where ``BlochCoeff`` is
        a symbolic representation of coefficients and a periodic function of ``k``.
        ``value`` can be scalar, array (both dense and sparse) or LinearOperator.
        This is accessible as a dict ``{BlochCoeff: value}``.

        Parameters
        ----------
        hamiltonian : Model, str, SymPy expression, dict or None (default)
            Symbolic representation of a Hamiltonian.  If a string, it is
            converted to a SymPy expression using ``kwant_continuum.sympify``.
            If a dict is provided, it should have the form
            ``{symbol: array}`` with all arrays the same size (dense or sparse).
            If symbol is not a BlochCoeff, it is passed through sympy.sympify,
            and should consist purely of a product of symbolic coefficients,
            no constant factors other than 1. `symbol` is then converted to BlochCoeff.
            `None` initializes a zero ``BlochModel``.
        locals : dict or ``None`` (default)
            Additional namespace entries for `~kwant_continuum.sympify`.  May be
            used to simplify input of matrices or modify input before proceeding
            further. For example:
            ``locals={'k': 'k_x + I * k_y'}`` or
            ``locals={'sigma_plus': [[0, 2], [0, 0]]}``.
        momenta : iterable of strings or Sympy symbols
            Names of momentum variables, default ``['k_x', 'k_y', 'k_z']`` or
            corresponding sympy symbols. Momenta are treated the same as other
            keys for the purpose of `keep`. Ignored when initialized with Model.
        keep : iterable of BlochCoeff (optional)
            Set of symbolic coefficients that are kept, anything that does not
            appear here is discarded. Useful for perturbative calculations where
            only terms to a given order are needed. By default all keys are kept.
            Ignored when initialized with Model.
        shape : tuple or None (default)
            Shape of the Model, must match the shape of all the values. If not
            provided, it is automatically found based on the shape of the input.
            Must be provided is ``hamiltonian`` is `None` or ``{}``. Empty tuple
            corresponds to scalar values. Ignored when initialized with Model.
        format : class or None (default)
            Type of the values in the model. Supported types are `np.complex128`,
            ``np.ndarray``, ``scipy.sparse.spmatrix`` and ``scipy.sparse.linalg.LinearOperator``.
            If ``hamiltonian`` is provided as a dict, all values must be of this type,
            except for scalar values, which are recast to ``np.complex128``.
            If ``format`` is not provided, it is inferred from the type of the values.
            If ``hamiltonian`` is not a dictionary, ``format`` is ignored and set to
            ``np.ndarray`` or ``hamiltonian.format`` if it is a ``Model``.
        """
        momenta = tuple(momenta)
        if hamiltonian is None:
            hamiltonian = {}
        if isinstance(hamiltonian, Model):
            # Use Model's init, only need to recast keys to BlochCoeff
            super().__init__(hamiltonian=hamiltonian.data,
                             locals=locals,
                             momenta=hamiltonian.momenta,
                             keep=hamiltonian.keep,
                             symbol_normalizer=lambda key: _bloch_normalizer(key, hamiltonian.momenta),
                             shape=hamiltonian.shape,
                             format=hamiltonian.format)
            # set these in case it was and empty Model
            self.format = hamiltonian.format
            self.shape = hamiltonian.shape
        elif isinstance(hamiltonian, abc.Mapping):
            keys = hamiltonian.keys()
            symbolic = all(not isinstance(k, BlochCoeff) for k in keys)
            hopping = all(isinstance(k, BlochCoeff) for k in keys)
            if hopping or hamiltonian == {}:
                # initialize as Model without any of the preprocessing
                super().__init__(hamiltonian,
                                 locals=locals,
                                 momenta=momenta,
                                 keep=keep,
                                 symbol_normalizer=lambda x: x,
                                 normalize=False,
                                 shape=shape,
                                 format=format,
                                )
            elif symbolic:
                # First cast it to model with restructuring, then try to interpret it as BlochModel
                self.__init__(Model(hamiltonian,
                                    locals=locals,
                                    momenta=momenta,
                                    keep=keep,
                                    normalize=True,
                                    shape=shape,
                                    format=format))
            else:
                raise ValueError('All keys must have the same type (sympy expression or BlochCoeff).')
        else:
            # Use Model to parse input
            self.__init__(Model(hamiltonian,
                                locals=locals,
                                momenta=momenta,
                                keep=keep,
                                shape=shape,
                                format=format))

    # Allow getting values using text keys
    def __getitem__(self, key):
        if key in self.data:
            return self.data[key]
        elif _bloch_normalizer(key, self.momenta) in self.data:
            return self.data[_bloch_normalizer(key, self.momenta)]
        else:
            return self.__missing__(key)

    def transform_symbolic(self, func):
        raise NotImplementedError('`transform_symbolic` is not implemented for `BlochModel`')

    def rotate_momenta(self, R):
        """Rotate momenta with rotation matrix R."""
        momenta = self.momenta
        assert len(momenta) == R.shape[0], (momenta, R)
        # do rotation on hopping vectors with transpose matrix
        R_T = np.array(R).astype(float).T
        result = self.zeros_like()
        result.data = {BlochCoeff(R_T @ hop, coeff): copy(val) for (hop, coeff), val in self.items()}
        return result

    def conj(self):
        """Complex conjugation."""
        result = self.zeros_like()
        result.data = {BlochCoeff(-hop, coeff.subs(sympy.I, -sympy.I)): val.conj()
                            for (hop, coeff), val in self.items()}
        return result

    def subs(self, *args, **kwargs):
        """Substitute symbolic expressions. See `Model.subs`."""
        model = self.tomodel(nsimplify=False)
        result = model.subs(*args, **kwargs)
        return BlochModel(result)

    def tosympy(self, nsimplify=False):
        """Return sympy representation of the Model.
        If nsimplify=True, attempt to rewrite numerical coefficients as exact formulas."""
        return self.tomodel(nsimplify=nsimplify).tosympy(nsimplify)

    def tomodel(self, nsimplify=False):
        """Convert to Model."""
        return Model({key.tosympy(self.momenta, nsimplify=nsimplify): copy(val)
                      for key, val in self.items()},
                     momenta=self.momenta,
                     keep={key.tosympy(self.momenta, nsimplify=nsimplify)
                                       for key in self.keep},
                     shape=self.shape,
                     format=self.format)


def _to_bloch_coeff(key, momenta):
    """Transform sympy expression to BlochCoeff if possible."""

    def is_hopping_expo(expo):
        # Check whether a sympy exponential represents a hopping.
        base, exponent = expo.as_base_exp()
        if base == e and any([momentum in exponent.atoms()
                              for momentum in momenta]):
            return True
        else:
            return False

    # We combine exponentials with the same base and exponent.
    key = sympy.powsimp(key, combine='exp')
    # Expand multiplication of brackets into sums.
    key = sympy.expand(key, power_base=False, power_exp=False,
                       mul=True, log=False, multinomial=False)
    if isinstance(key, sympy.add.Add):
        raise ValueError("Key cannot be a sum of terms.")
    # Key is a single exponential.
    if isinstance(key, sympy.power.Pow):
        base, exp = key.as_base_exp()
        # If the exponential is a hopping, store it
        # with coefficient 1.
        if is_hopping_expo(key):
            hop_expo = key
            coeff = sympy.numbers.One()
        # If it is not a hopping, it belongs to the coeff.
        else:
            hop, coeff, hop_expo = np.zeros((len(momenta,))), key, None
    # Key is the product of an exponential and some extra stuff.
    elif sympy.power.Pow in [type(arg) for arg in key.args]:
        # Check that a natural exponential is present, which also
        # includes momenta in its arguments.
        # First find all exponentials.
        find_expos = [ele for ele in key.args if ele.is_Pow]
        # Then pick out exponentials that are hoppings.
        hop_expos = [expo for expo in find_expos if is_hopping_expo(expo)]
        # We should find at most one exponential that represents a
        # hopping, because all exponentials with the same base have been
        # combined.
        if len(hop_expos) == 1:
            hop_expo = hop_expos[0]
            coeff = sympy.simplify(key / hop_expo)
        # If none of the exponentials match the hopping structure, the
        # exponentials that are present are parts of the coefficient,
        # so this is an onsite term.
        elif not len(hop_expos):
            hop, coeff, hop_expo = np.zeros((len(momenta,))), key, None
        # Should never be called.
        else:
            raise ValueError("Unable to read the hoppings in "
                             "conversion to BlochCoeff.")
    # If the key contains no exponentials, then it is not a hopping.
    else:
        hop, coeff, hop_expo = np.zeros((len(momenta,))), key, None
    # Extract hopping vector from exponential
    # If the exponential contains more arguments than the hopping,
    # append it to coeff.
    if hop_expo is not None:
        base, exponent = hop_expo.as_base_exp()
        if base != e or type(exponent) not in (sympy.Mul, sympy.Add):
            raise ValueError('Incorrect format of exponential.')
        # Pick out the real space part, remove the complex i,
        # expand any brackets if present.
        arg = exponent.expand()
        # Check that the momenta all have i as a prefactor
        momenta_present = [momentum for momentum in momenta
                           if momentum in arg.atoms()]
        if not all(
            [sympy.I in (arg.coeff(momentum)).atoms()
             for momentum in momenta_present]
        ):
            raise ValueError(
                "Momenta in hopping exponentials should have a complex prefactor."
            )
        hop = [sympy.expand(arg.coeff(momentum)/sympy.I)
               for momentum in momenta]
        # We do not allow sympy symbols in the hopping, should
        # be numerical values only.
        if any([isinstance(item, sympy.symbol.Symbol)
                        for ele in hop for item in ele.atoms()
                        if isinstance(ele, sympy.Expr)]):
            raise ValueError(
                "Real space part of the hopping must be numbers, not symbols."
            )
        # If the exponential contains something extra other than the
        # hopping part, we append it to the coefficient.
        spatial_arg = sympy.I*sum([ele*momentum for ele, momentum in zip(momenta, hop)])
        diff = sympy.nsimplify(sympy.expand(arg - spatial_arg))
        coeff = sympy.simplify(coeff * e**diff)
        hop = np.array(hop).astype(float)
    # Make sure there is no momentum dependence in the coefficient.
    if any([momentum in coeff.atoms() for momentum in momenta]):
        raise ValueError(
            "All momentum dependence should be confined to hopping exponentials."
        )
    return BlochCoeff(hop, coeff)


@lru_cache()
def _find_momenta(momenta):
    if any(isinstance(i, int) for i in momenta):
        raise TypeError('Momenta should be strings or sympy symbols.')
    elif all(m in _commutative_momenta for m in momenta):
        return tuple(momenta)
    else:
        _momenta = [kwant_continuum.sympify(k) for k in momenta]
        return tuple(kwant_continuum.make_commutative(k, k)
                        for k in _momenta)


@lru_cache(maxsize=1000)
def _symbol_normalizer(key):
    return sympy.expand_power_exp(sympy.sympify(key))


@lru_cache(maxsize=1000)
def _bloch_normalizer(key, momenta):
    if isinstance(key, BlochCoeff):
        return key
    else:
        return _to_bloch_coeff(key, momenta)


def _shape_and_format(val):
    # Find shape and type of val
    format = type(val)
    try:
        shape = val.shape
    except AttributeError:
        # Treat it as a scalar
        shape = ()
    if issubclass(format, Number):
        # Cast all numbers to np.complex128
        format = np.complex128
    elif issubclass(format, np.ndarray):
        format = np.ndarray
    elif issubclass(format, scipy.sparse.linalg.LinearOperator):
        # Make all subclasses of LinearOperator work
        format = scipy.sparse.linalg.LinearOperator
    elif not issubclass(format, scipy.sparse.spmatrix):
        raise ValueError('Only `formats` which are subclasses of `np.ndarray`, `scipy.sparse.spmatrix` '
                         '`scipy.sparse.linalg.LinearOperator` or `Number` are supported.')
    return shape, format



import numpy as np
import scipy.linalg as la
import scipy.sparse
import itertools as it
from copy import deepcopy

from .linalg import matrix_basis, nullspace, split_list, simult_diag, commutator, \
                    prop_to_id, sparse_basis, mtm, family_to_vectors, solve_mat_eqn, \
                    allclose
from .model import Model, BlochModel, BlochCoeff
from .groups import PointGroupElement, ContinuousGroupGenerator, generate_group, \
                    set_multiply, L_matrices, spin_rotation, time_reversal, \
                    particle_hole, chiral, inversion, rotation, mirror, PrettyList

from . import kwant_continuum
from .kwant_linalg_lll import lll, cvp, voronoi

### Top level function for symmetry finding

def symmetries(model, candidates=None, continuous_rotations=False,
               generators=False, prettify=False, num_digits=8, verbose=False,
               sparse_linalg=False):
    """
    Find symmetries of the Hamiltonian family described by model.

    Parameters:
    -----------
    model : Model
        Model or BlochModel which represents family of Hamiltonians. Every
        symbolic prefactor is treated as a free parameter, and model.momenta
        as independent momentum variables.
    candidates : iterable of PointGroupElements or None
        Set of candidate PointGroupElements used for finding discrete
        symmetries. Must have .U attribute set to None.
        It is advised that candidates forms a group, as combinations of
        elements in candidates are not tested.
        If model is a Model that describes a Bloch Hamiltonian, the rotation
        matrices must be either integer or Sympy matrices, as exact arythmetic
        is assumed. Floating point rotation matrices are acceptable for k.p
        Models and BlochModels.
        If None (default), only discrete onsite symmetries (time reversal,
        particle-hole, chiral) are found.
    continuous_rotations : bool (default False)
        Whether to search for continuous rotation symmetries.
    generators : bool (default False)
        If True, only a set of generators are returned, otherwise
        the full discrete symmetry group is returned.
    prettify : bool (default False)
        Whether to carry out sparsification of the continuous symmetry
        generators, in general an arbitrary linear combination of the
        symmetry generators is returned.
    num_digits : float
        Absolute precision when deciding whether symmetry leaves Hamiltonian
        invariant and prettifying the result with prettify=True.
    verbose : bool
        Whether to print additional information.
    sparse_linalg : bool
        Whether to use sparse linear algebra in the calculation.
        Can give large performance gain in large systems.

    Returns:
    --------
    disc_sym : list of PointGroupElement
        Discrete symmetries of model, always a group generated by a subset of
        candidates. If generators=False, it is a closed group, if generators=True
        it is a (not necessarily minimal) generator set, may be empty.
    cont_sym : list of ContinuousGroupGenerator
        List of linearly independent continuous symmetry generators, may be
        empty. Onsite conserved quantities are listed first, then continuous
        rotation generators. The trivial conserved quantity proportional to
        identity is not included.
    """
    if not issubclass(model.format, (np.ndarray, scipy.sparse.spmatrix)):
        raise ValueError('Symmetry finding is only supported for Models with '
                         '`np.ndarray` or `scipy.sparse.spmatrix` data type.')

    # Find onsite conserved quantites and projectors onto blocks.
    Ps = _reduce_hamiltonian(np.array(list(model.values())), sparse=sparse_linalg)
    cont_sym = conserved_quantities(Ps, prettify=prettify, num_digits=num_digits)

    # Find continuous rotations
    if continuous_rotations:
        cont_sym += continuous_symmetries(model, Ps=Ps, prettify=prettify,
                                          num_digits=num_digits, sparse_linalg=sparse_linalg)

    # Find discrete symmetries
    if candidates is None:
        # Make discrete onsite symmetries
        dim = len(model.momenta)
        candidates = generate_group({time_reversal(dim), particle_hole(dim), chiral(dim)})

    if candidates:
        disc_sym, _ = discrete_symmetries(model, set(candidates), Ps=Ps,
                                          generators=generators, verbose=verbose,
                                          sparse_linalg=sparse_linalg)
        disc_sym = sorted(list(disc_sym))
    else:
        disc_sym = []

    return PrettyList(disc_sym), PrettyList(cont_sym)


### Lie Algebra utility functions

def struct_const(gens):
    # Calculate structure constants of Lie algebra given by 'gens'

    def expand(v, bas):
        # expand v in incomplete nonorthogonal basis bas
        c, res, _, _ = la.lstsq(bas, v)
        if not np.all(np.isclose(res, 0)):
            raise ValueError('Vector outside of space spanned by basis!')
        return c

    bas = np.array([L.flatten() for L in gens]).T
    return np.array(
        [
            [
                expand(commutator(L1,L2).flatten(), bas)
            for L2 in gens]
        for L1 in gens])


def killing(gens):
    # Calculate Killing form
    C = struct_const(gens)
    # 'ade,bed->ab'
    return np.tensordot(C, C, axes=((1, 2), (2, 1)))


def casimir(gens):
    # Calculate quadratic Casimir operator
    g = killing(gens)
    if np.isclose(la.det(g), 0):
        raise ValueError('Lie-algebra not semisimple!')
    # 'ab,aik,bkj->ij'
    return np.tensordot(la.inv(g), gens.dot(gens), axes=((0, 1), (0, 2)))
    # Generalized to the case when includes a commutative subalgebra ???
    # return np.einsum('ab,aik,bkj->ij',la.pinv(g),gens,gens)


def separate_lie_algebra(gens):
    """
    Separate Lie-algebra defined by the list of linearly
    independent generators 'gens' into its center and semisimple parts.

    Parameters:
    -----------------
    gens : ndarray
        3 index array of shape (m, N, N), Lie-algebra generators

    Returns:
    ---------------
    gensc : ndarray
        generators of the center, shape (mc, N, N)
    genss : ndarray
        generators of semisimple part, shape (ms, N, N)
    """
    g = killing(gens)
    c, s = nullspace(g, return_complement=True)
    # 'ab,bij->aij'
    gensc = np.tensordot(c, gens, axes=(1, 0))
    genss = np.tensordot(s, gens, axes=(1, 0))
    return gensc, genss


def symmetry_adapted_sun(gens, check=False):
    r"""
    Find symmetry adapted basis of the simple 'su(d)'
    Lie-algebra representation defined by generators 'gens'.
    It is assumed that the representation is the direct sum
    of 'n' identical 'su(d)' representations, i.e. in the symmetry
    adapted basis the generators have the form 'L \otimes 1_{nxn}' where
    'L' runs over a generator set of all 'd*d' Hermitian matrices.

    Parameters:
    -----------------
    gens : ndarray
        3 index array of shape '(d**2-1, n*d, n*d)' with 'n' and 'd' integers,
        list of 'su(d)' generators.

    Returns:
    ---------------
    Ps : ndarray
        3 index array of shape '(d, n*d, n)', list of projectors onto the symmetry
        adapted subspaces. 'np.einsum('aij,ab,bkj->ik', Ps, L, Ps.conjugate()))'
        is a symmetry spanned by 'gens' for any 'd*d' Hermitian matrix 'L'.
        Stacking these produces the unitary transformation 'U = np.hstack(Ps)' to the
        symmetry adapted basis.
    check : bool
        Whether to check the final result.
    """
    d = np.sqrt(len(gens) + 1)
    n = gens.shape[-1] / d
    if not (n.is_integer() and d.is_integer()):
        raise ValueError('Shape of gens is incompatible with it being direct sum of'
                         'n identical copies of su(d) representation.')
    n, d = int(n), int(d)
    # Trivial case when it is a full SU(N)
    if n == 1:
        return np.array(np.split(np.eye(n*d), n*d, 1))

    # Iteratively split the irreps. Goal is to find a basis in the n*d dimensional
    # Hilbert-space, where the n vectors belonging to the identity factor in the symmetry
    # adapted basis are grouped together. A degenerate eigensubspace of a generator has
    # dimension of a multiple of n. If it is exactly n, we can be sure that we found
    # such a subspace. If it is f*n, it is the union of f such subspaces. We project the
    # remaining generators in this subspace, they will still span all matrices with the
    # structure L \otimes 1_{nxn} with L f*f. We pick a new generator and repeat the process.
    # In the generic case (a random combination of the generators) the first generator
    # already splits all the irreps, but in the standard matrix basis this is not the case.
    unsplit = [np.eye(n*d)]
    split = []
    for g in gens:
        Ps = unsplit
        unsplit = []
        for P in Ps:
            # project generator in unsplit subspace, rest of the generators still
            # span the full space of Hermitian matrices in the restricted space
            gr = P.T.conj() @ g @ P
            evals, U = la.eigh(gr)
            # find the degenerate eigenspaces
            blocks = split_list(evals)
            for b, e in blocks:
                if e - b == n:
                    # if n long degenerate block, it is split
                    split.append(P @ U[:, b:e])
                else:
                    assert (e - b) % n == 0
                    # if it is a multiple of n, put it back in unsplit
                    unsplit.append(P @ U[:, b:e])
        if unsplit == []:
            break
    else:
        # we could not split all the subspaces
        raise ValueError('Algorithm failed, likely not and su(d) representation')
    Ps = np.array(split)

    P0 = Ps[0]
    for i, P in enumerate(Ps[1:]):
        # Adjust the basis of every block to the basis of the first block,
        # such that all the generators are proportional to the identity
        # in every block.
        for g in gens:
            # find a generator where the block is nonzero
            gblock = P0.T.conj() @ g @ P
            if np.allclose(gblock, 0):
                continue
            # then it is proportional to unitary, transform it
            # to proportional to identity
            u, s, vh = la.svd(gblock)
            U = (((u @ vh).T).conj())
            Ps[i+1] = Ps[i+1] @ U
            break
        else:
            # we could not fix some of the bases
            raise ValueError('Algorithm failed, likely not and su(d) representation')
    if check:
        # check that every block is proportional to the identity
        U = np.hstack(Ps)
        genst = mtm(U.T.conjugate(), gens, U)
        assert np.all([[prop_to_id(genst[i, n*j1:n*(j1+1), n*j2:n*(j2+1)])[0]
                     for j1, j2 in it.combinations_with_replacement(range(d), 2)]
                     for i in range(len(genst))])
    return Ps


### Continuous onsite symmetry finding

def _reduce_hamiltonian(H, sparse=False):
    """
    Find the unitary symmetries and the symmetry adapted basis of a family
    of matrices H. In the symmetry adapted basis the matrices are
    simultaneously block-diagonalized, each block corresponding to an
    irreducible subspace.

    Parameters:
    -----------------
    H : ndarray
        3 index array of shape (m, N, N), or list of NxN Hermitian matrices
        defining the family.
    sparse : bool
        Whether to use sparse linear algebra in the calculation.
        Can give large performance gain in large systems.

    Returns:
    ---------------
    Preduced : tuple of ndarrays
        symmetry adapted bases (Pr_1, Pr_2, ...) where Pr_i is a 3 index array of
        shape (d_i, N, n_i) with sum_i n_i * d_i = N, each have the same format
        as the output of symmetry_adapted_sun and satisfies that
        `np.einsum('aij,ab,bkj->ik', Pr_i, L, Pr_i.conjugate()))` is a symmetry
        in the i'th block for any d_i x d_i Hermitian matrix `L`. Each Pr_i
        corresponds to d_i irreducible subspaces of n_i dimensions, in which
        the H's act identically.
    """
    # Find all symmetry generators
    gens = solve_mat_eqn(H, hermitian=True, traceless=True, sparse=sparse)

    if len(gens) == 0:
        return (np.array([np.eye(H[0].shape[-1])]),)
    # Find the center
    gensc, _ = separate_lie_algebra(gens)
    if len(gensc) == 0:
        Ps = [np.eye(H[0].shape[-1])]
    else:
        # Simultaneaously diagonalize central generators
        Ps = simult_diag(gensc)

    Preduced = tuple()
    # Find symmetry adapted basis of restricted LA in each subspace
    for P in Ps:
        # Suppot list of sparse matrices
        Hr = [P.T.conjugate() @ h @ P for h in H]
        gensr = solve_mat_eqn(Hr, hermitian=True, traceless=True, sparse=sparse)
        # Find symmetry adapted basis in subspace
        if len(gensr) == 0:
            P2s = [np.eye(Hr[0].shape[-1])]
        else:
            P2s = symmetry_adapted_sun(gensr)
        Preduced += (np.array([np.dot(P, P2i) for P2i in P2s]),)

    return Preduced


def conserved_quantities(Ps, prettify=False, num_digits=3):
    """
    Construct a full set of conserved quantities from the projectors.

    Parameters:
    -----------
    Ps : list fo 3 index ndarrays
        projectors 'Ps' returned '_reduce_hamiltonian()'
    prettify : bool
        If true it finds a nice sparse basis of the conserved
        quantities, they are generally returned in a random basis, but
        any linear combination is also conserved.

    Returns:
    --------
    list of ContinuousGroupGenerators
        conserved quantities that all commute with the family of
        Hamiltonians. The identity matrix is excluded.
    """
    Ls = []
    # Iterate over symmetry blocks
    for i, P in enumerate(Ps):
        # generate basis of Hermitian matrices with subblock size
        # First block does not have identity, so the identity is not
        # among the conserved quantities
        bas = matrix_basis(P.shape[0], traceless=(i==0))
        for l in bas:
            # construct conserved L that acts with l on all the subblocks
            # of the original space at the same time (sum over j)
            # 'aij,ab,bkj->ik'
            L = np.tensordot(np.tensordot(P, l, axes=((0), (0))), P.conj(),
                             axes=((2, 1), (0, 2)))
            # Make it traceless
            Ls.append(L - np.trace(L) / len(L) * np.eye(len(L)))
    Ls = np.array(Ls)
    # Sparsify the matrices using reduced row echelon form
    if len(Ls) > 1 and prettify:
        Lsf = Ls.reshape(Ls.shape[0], -1)
        Lsf = sparse_basis(Lsf, reals=True, num_digits=num_digits)
        Ls = Lsf.reshape(Lsf.shape[0], *Ls.shape[1:])
    return PrettyList([ContinuousGroupGenerator(None, L) for L in Ls])


### Point group symmetry finding

def discrete_symmetries(model, candidates, Ps=None, generators=False,
                        verbose=False, sparse_linalg=False):
    """Find point group symmetries of Hamiltonians family.
    Optimized version to reduce number of tests,
    uses sympy exact rotation matrices

    Parameters:
    -----------
    model : Model
        Model which represents family of Hamiltonians
    candidates : set of PointGroupElements
        Set of candidate PointGroupElements. Must have .U attribute
        set to None.
    Ps : ndarray, optional
        Projectors as returned by _reduce_hamiltonian.
    generators : bool
        If true, only a set of generators are returned, otherwise
        the full symmetry group is returned.
    sparse_linalg : bool
        Whether to use sparse linear algebra in the calculation.
        Can give large performance gain in large systems.
    verbose : bool

    Returns:
    --------
    genset or symset : set of PointGroupElement
        Symmetries of model.
    ### TODO: remove Ps from return
    Ps : ndarray
        Projectors as returned by _reduce_hamiltonian.
    """
    if not issubclass(model.format, (np.ndarray, scipy.sparse.spmatrix)):
        raise ValueError('Symmetry finding is only supported for Models with '
                         '`np.ndarray` or `scipy.sparse.spmatrix` data type.')
    symmetry_candidates = deepcopy(candidates)
    m = len(symmetry_candidates)
    # Reduce Hamiltonian by onsite unitaries
    if not Ps:
        Ps = _reduce_hamiltonian(list(model.values()), sparse=sparse_linalg)
    # After every step, symlist is guaranteed to form a group, start with the trivial group
    e = next(iter(candidates)).identity()
    e.U = np.eye(Ps[0].shape[1])
    # set of PointGroupElements
    symset = {e}
    # set of generators
    genset = set()
    symmetry_candidates -= symset
    not_symmetries = set()
    n = 0
    while symmetry_candidates:
        # For reproducibility, iterate over elements in sorted order
        # instead of simply popping an arbitrary element
        g = min(symmetry_candidates)
        symmetry_candidates -= {g}
        # Find unitary part
        gr = _find_unitary(model, Ps, g, sparse=sparse_linalg)
        if gr.U is not None:
            # Check that it's indeed a symmetry
            assert gr.apply(model) == model, (n, gr)
            genset.add(gr)
            # Needless to test anything in the group generated by the
            # symmetries found already, they are symmetries for sure.
            symset = generate_group({gr} | genset)
            symmetry_candidates -= symset
            # Needless to test anything of the form Q*R, Q*R**-1,
            # R*Q, R**-1*Q where Q is a symmetry and R is not,
            # it is surely not a symmetry.
            # higher powers of R may still be symmetries.
            new_ns = set_multiply({g, g.inv()}, not_symmetries)
            new_ns -= not_symmetries
            new_ns |= set_multiply(new_ns, {g, g.inv()})
            not_symmetries |= new_ns
        else:
            new_ns = set_multiply({g, g.inv()}, symset)
            new_ns -= not_symmetries
            new_ns |= set_multiply(symset, new_ns)
            not_symmetries |= new_ns
        symmetry_candidates -= not_symmetries
        n+=1
    if verbose:
        print('{} symmetries explicitely tested of {} candidates.'.format(n, m))
    assert not any(g.U is None for g in symset)
    if generators:
        return genset, Ps
    else:
        return symset, Ps


def _find_unitary(model, Ps, g, sparse=False, checks=False):
    """Test if the candidate k-space symmetry is (anti)unitary (anti)symmetry,
    if not, return 'None', if yes, return the unitary part of the transformation
    'U'. Checked condition if unitary: U H(inv(R) k) = (+/-) H(k) U
    Checked condition if antiunitary: U H(-inv(R) k).conj() = (+/-) H(k) U.
    RHS (+/-) stands for symmetry/antisymmetry.

    Parameters:
    -----------
    model : Model
        model which represents family of Hamiltonians
    Ps : iterable of ndarrays
        Projectors onto the irreducible subspaces of on-site symmetries,
        as returned by '_reduce_hamiltonian'
    g : PointGroupElement
        Standard representation of symmetry operator. g.U must be None.
    checks : bool
        Whether to perform checks.

    Returns:
    --------
    gr : PointGroupElement
        Point group operator with gr.U set to the Hilbert space action
        of the symmetry is found, otherwise identical to g.
    """
    if not issubclass(model.format, (np.ndarray, scipy.sparse.spmatrix)):
        raise ValueError('Symmetry finding is only supported for Models with '
                         '`np.ndarray` or `scipy.sparse.spmatrix` data type.')
    if g.U is not None:
        raise ValueError('g.U must be None.')
    Rmodel = g.apply(model)
    if set(model) != set(Rmodel):
        return g
    HR, HL = [], []
    # Only test eigenvalues if all matrices are Hermitian
    ev_test = True
    for key, matL in model.items():
        HR.append(matL)
        matR = Rmodel[key]
        HL.append(matR)
        ev_test = ev_test and allclose(matL, matL.T.conj()) and allclose(matR, matR.T.conj())
    # Need to carry conjugation on left side through P
    if g.conjugate:
        PsL = [P.conj() for P in Ps]
    else:
        PsL = Ps
    HRs = [[P[0].T.conj() @ hR @ P[0] for hR in HR] for P in Ps]
    HLs = [[P[0].T.conj() @ hL @ P[0] for hL in HL] for P in PsL]

    squares_to_1 = g * g == g.identity()
    block_dict = _find_unitary_blocks(HLs, HRs, Ps, conjugate=g.conjugate, ev_test=ev_test,
                                      squares_to_1=squares_to_1, sparse=sparse)
    S = _construct_unitary(block_dict, Ps, conjugate=g.conjugate, squares_to_1=squares_to_1)

    if checks:
        HR, HL = np.array(HR), np.array(HL)
        for i, j in it.product(range(len(Ps)), range(len(Ps))):
            for a, b in it.product(range(len(Ps[i])), range(len(Ps[j]))):
                if i !=j or a!=b:
                    assert np.allclose(mtm(PsL[i][a].T.conj(), HL, PsL[j][b]), 0)
                    assert np.allclose(mtm(Ps[i][a].T.conj(), HR, Ps[j][b]), 0)
                else:
                    assert allclose(mtm(PsL[i][a].T.conj(), HL, PsL[j][b]), HLs[i])
                    assert allclose(mtm(Ps[i][a].T.conj(), HR, Ps[j][b]), HRs[i])
        if (not g.conjugate) and (not g.antisymmetry) and (S is not None):
            assert allclose(S @ HL, HR @ S)

    return PointGroupElement(g.R, g.conjugate, g.antisymmetry, S)


def _find_unitary_blocks(HLs, HRs, projectors, squares_to_1=True, conjugate=False,
                         ev_test=True, sparse=False):
    """Find candidate symmetry linear spaces in all blocks.
    HLs and HRs are lists of reduced Hamiltonians (families) that go to left and right side
    of the equations.

    Returns a dictionary {(i, j): Uij} of all symmetry candidate blocks that have a
    nonzero solution of Uij @ HLs[j] = HRs[i] @ Uij for Uij.

    If squares_to_1=True, it is assumed that the operators square is proportional to 1
    in every block. The search is limited to j <= i, the diagonal blocks have a phase choice
    and the off-diagonal blocks with j > i are constructed to ensure squaring to +-1.
    Otherwise the blocks Uij and Uji are calculated independently.

    If ev_test=True the eigenvalues of the matrices are tested first
    """
    # Only need to find symmetries in half of each block of the Hamiltonian.
    # We take blocks in the lower triangular half and on the diagonal.
    block_dict = {}
    ind = range(len(projectors))
    # Pretest eigenvalues
    if ev_test:
        evsL = [[la.eigvalsh(h) for h in HLs[i]] for i in ind]
        evsR = [[la.eigvalsh(h) for h in HRs[i]] for i in ind]
    for (i, j) in it.product(ind, ind):
        # Only do j <= i if squares to 1
        if squares_to_1 and j>i:
            continue
        # Only allowed between blocks of identical shape
        if projectors[i].shape != projectors[j].shape:
            continue
        # Pretest eigenvalues
        if ev_test:
            if not allclose(evsL[j], evsR[i]):
                continue
        # Find block ij of the symmetry operator
        block_dsymm = solve_mat_eqn(HLs[j], HRs[i], hermitian=False, traceless=False, sparse=sparse, k_max=2)
        # Normalize block_dsymm such that it is close to unitary. The matrix
        # returned by solve_mat_eqn is normalized such that Tr(X.T.conj() @ X) is close to 1.
        block_dsymm = np.sqrt(block_dsymm.shape[-1]) * block_dsymm
        # If the space is not empty, we store it and the indices of the block.
        if len(block_dsymm):
            # There should be only one solution, which is invertible
            if len(block_dsymm) > 1 or np.isclose(la.det(block_dsymm[0]), 0):
                raise ValueError('Hamiltonian blocks have residual symmetry.')
            block_dsymm = block_dsymm[0]
            assert allclose(block_dsymm @ HLs[j],
                               HRs[i] @ block_dsymm)
            # The block must be proportional to a unitary
            prop_to_I, coeff = prop_to_id(block_dsymm.dot(block_dsymm.T.conj()))
            assert prop_to_I and np.isclose(np.imag(coeff), 0) and np.real(coeff)>0
            # Normalize such that it is unitary
            block_dsymm = block_dsymm/np.sqrt(coeff)
            block_dict[(i, j)] = block_dsymm
            # If squares to 1, fill out the lower triangle
            if squares_to_1:
                block_dict[(i, j)], block_dict[(j, i)] = _nice_square(block_dsymm, (i == j), conjugate)
    return block_dict


def _nice_square(block_dsymm, diagonal, conjugate):
    # Make sure blocks square to +-1
    # Diagonal blocks need proper phase choice if unitary,
    # nothing to be done if antiunitary, must square to +-1
    if diagonal:
        if conjugate:
            prop_to_I, coeff = prop_to_id(block_dsymm.dot(block_dsymm.conj()))
            assert prop_to_I and (np.isclose(coeff, 1) or np.isclose(coeff, -1))
        else:
            prop_to_I, coeff = prop_to_id(block_dsymm.dot(block_dsymm))
            assert prop_to_I and np.isclose(np.abs(coeff), 1)
            block_dsymm = block_dsymm/np.sqrt(coeff)
        return block_dsymm, block_dsymm
    # Off-diagonal blocks are chosen such that it squares to +1
    else:
        if conjugate:
            return block_dsymm, block_dsymm.T
        else:
            return block_dsymm, block_dsymm.T.conj()


def _construct_unitary(block_dict, projectors, conjugate=False, squares_to_1=True):
    """Search for combinations of blocks of the symmetry operator that when combined give a symmetry
    in canonical form, i.e. with only one nonzero block per row and column, and attempt to construct
    the operator. """
    block_keys = block_dict.keys()
    n = len(projectors)
    # Need to find a canonical form of the symmetry operator.
    # Iterate over all combinations of the nonzero blocks
    for perm in it.combinations(block_keys, len(projectors)):
        # Make the corresponding matrix
        M = np.zeros((n, n))
        for i, j in perm:
            M[i, j] = 1
        # Check that it is a permutation matrix
        if not np.all(M @ M.T == np.eye(n)):
            continue
        # If squares_to_1 check that (j, i) block is also present
        if squares_to_1 and not np.all(M == M.T):
            continue
        # Construct the symmetry operator and return
        # Initialize complete symmetry operator
        S = np.zeros((projectors[0].shape[-2], projectors[0].shape[-2]), dtype=complex)
        # Iterate over all blocks that give a canonical form
        for i, j in perm:
            block = block_dict[(i, j)]
            # Rebuild full operator using projectors
            pi, pj = projectors[i], projectors[j]
            # Use conjugate projector if antiunitary
            if conjugate:
                # 'aij, jk, alk -> il'
                S += np.tensordot(pi @ block, pj, axes=((0, 2), (0, 2)))
            else:
                S += np.tensordot(pi @ block, pj.conj(), axes=((0, 2), (0, 2)))
        assert prop_to_id(S.dot(S.T.conj()))[0]
        return S
    # If we cannot construct a canonical symmetry operator, return None
    return None

### Continuous spatial symmetry finding

def continuous_symmetries(model, Ps=None, prettify=True, num_digits=8, sparse_linalg=False):
    """Find continuous rotation symmetries of Hamiltonian family represented
    by model. Hamiltonian is reduced, so on-site continuous symmetries
    are factored out.

    Parameters
    ----------
    model : Model
        symbolic representation of the Hamiltonian family
    Ps : ndarray, optional
        Projectors as returned by _reduce_hamiltonian
    prettify : bool
        Whether to carry out sparsification of the results, in general an
        arbitrary linear combination of the symmetry generators is returned.
    num_digits : float
        Absolute precision when prettifying the result.
    sparse_linalg : bool
        Whether to use sparse linear algebra in the calculation.
        Can give large performance gain in large systems.

    Returns
    -------
    symmetries : list of ContinuousGroupGenerator
        List of linearly independent symmetry generators.
    """
    if not issubclass(model.format, (np.ndarray, scipy.sparse.spmatrix)):
        raise ValueError('Symmetry finding is only supported for Models with '
                         '`np.ndarray` or `scipy.sparse.spmatrix` data type.')
    if isinstance(model, BlochModel):
        # BlochModel cannot have continuous rotation symmetry
        return []
    if not Ps:
        Ps = _reduce_hamiltonian(list(model.values()), sparse=sparse_linalg)
    reduced_hamiltonians = _reduced_model(model, Ps)
    dim = len(model.momenta)
    if dim <= 1:
        # There is no continuous rotation in 0 and 1 D
        return []
    Rs = lambda: matrix_basis(dim, antihermitian=True, real=True)
    # length of Rs
    NR = dim * (dim - 1) / 2
    blockdims = [list(rham.values())[0].shape[0] for rham in reduced_hamiltonians]
    # Blocks corresponding to real space rotations
    Rblocks = []
    # Blocks corresponding to Hilbert space transformations
    Lblocks = []
    # Iterate over all the reduced hamiltonian blocks
    for rham in reduced_hamiltonians:
        blockdim = rham.shape[0]
        L = None
        # Generate all reduced hamiltonians transformed by spatial part
        trf_hams = [ContinuousGroupGenerator(1j * R, L).apply(rham) for R in Rs()]
        # Generate all keys that appear as transformed reduced hamiltonians
        keys = rham.keys()
        for th in trf_hams:
            keys |= th.keys()
        keys = list(keys)

        # Iterate over all reduced hamiltonians transformed by only spatial part
        Rblock = family_to_vectors(trf_hams, all_keys=keys).T
        Rblocks.append(Rblock)

        # Iterate over all reduced hamiltonians transformed by only hilbert space part
        R = None
        Ls = matrix_basis(blockdim, traceless=True)
        trf_hams = [ContinuousGroupGenerator(R, L).apply(rham) for L in Ls]
        Lblock = family_to_vectors(trf_hams, all_keys=keys).T
        Lblocks.append(Lblock)

    # Build constraint matrix: first the spatial blocks in a column, then the
    # Hilbert-space blocks as block-diagonal, as they do not mix different
    # reduced subspaces.
    constraints = np.hstack(((np.vstack(Rblocks), la.block_diag(*Lblocks))))

    # Find the linearly independent solutions
    null_vecs = nullspace(constraints, sparse=sparse_linalg)
    if prettify:
        null_vecs = sparse_basis(null_vecs, reals=False, num_digits=num_digits)

    # Build the symmetry generator
    symmetries = []
    for v in null_vecs:
        # Build spatial part
        R = sum(1j * R * v[i] for i, R in enumerate(Rs()))
        # Build unitary part for each block
        L = np.zeros((Ps[0].shape[1], Ps[0].shape[1]), dtype=complex)
        for i, rham in enumerate(reduced_hamiltonians):
            # There is no action in a 1D block
            if blockdims[i] > 1:
                Ls = matrix_basis(blockdims[i], traceless=True)
                blockind = int(NR + np.sum([d**2-1 for d in blockdims[:i]]))
                l = sum(l * v[blockind + j] for j, l in enumerate(Ls))
                L += np.einsum('aij,jl,akl->ik', Ps[i], l, Ps[i].conj())
        g = ContinuousGroupGenerator(R, L)
        symmetries.append(g)
        # Check that it is a symmetry
        trf = g.apply(model)
        assert trf.allclose(0, atol=1e-6), (trf, g)
    return PrettyList(symmetries)


def _reduced_model(model, Ps=None):
    """
    Construct reduced Hamiltonians in monomial form.

    Parameters
    ----------
    model : Model
        symbolic representation of the Hamiltonian family
    Ps : list fo 3 index ndarrays
        projectors 'Ps' returned '_reduce_hamiltonian()'
        Optional, can be provided to speed up the calculation.

    Returns
    -------
    reduced_hamiltonians : list of Model
        List of reduced Hamiltonian families, each projected
        on the symmetry irreducible subspaces.
    """
    if not issubclass(model.format, (np.ndarray, scipy.sparse.spmatrix)):
        raise ValueError('Symmetry finding is only supported for Models with '
                         '`np.ndarray` or `scipy.sparse.spmatrix` data type.')
    if Ps is None:
        Ps = _reduce_hamiltonian(np.array([H for H in model.values()]))
    reduced_hamiltonians = []
    for P in Ps:
        Hr = P[0].T.conj() @ model @ P[0]
        reduced_hamiltonians.append(Hr)
    return reduced_hamiltonians

### Bravais lattice symmetry finding

def bravais_point_group(periods, tr=False, ph=False, generators=False, verbose=False):
    """Find the  point group of the Bravais-lattice defined by periods.

    Parameters:
    -----------
    periods: array
        Translation vectors as row vectors, arranged into a 2D array.
    tr, ph: bool (default False)
        Whether to return time reversal and particle-hole operators.
        If False, only pure point-group operators are returned.
    generators: bool (default False)
        If True only a (not necessarily minimal) generator set of the
        symmetry group is returned.
    verbose: bool (default False)
        If True, the name of the Bravais lattice is printed.

    Returns:
    --------
    set of PointGroupElements
    """
    periods = np.asarray(periods)
    dim = periods.shape[0]
    # Project onto the subspace spanned by the translations
    if periods.shape[1] > periods.shape[0]:
        proj, r = la.qr(periods.T, mode='economic')
        sign = np.diag(np.diag(np.sign(r)))
        proj = sign @ proj.T
        periods = periods @ proj.T

    # get lll reduced basis
    periods, _ = lll(periods)
    # get nearest neighbors
    neighbors = voronoi(periods, reduced=True, rtol=1e-5) @ periods
    neighbors = neighbors[:len(neighbors)//2]
    num_eq, sets_eq = equals(neighbors)

    # Everey Bravais-lattice group contains inversion
    gens = {inversion(dim)}

    if dim==1:
        # The only bravais lattice group in 1D only contains inversion
        pass
    elif dim==2:
        gens |= bravais_group_2d(neighbors, num_eq, sets_eq, verbose)
    elif dim==3:
        gens |= bravais_group_3d(neighbors, num_eq, sets_eq, verbose)
    else:
        raise NotImplementedError('Only 1, 2, and 3 dimensional translation symmetry supported.')

    if tr:
        TR = time_reversal(dim)
        gens.add(TR)
    if ph:
        PH = particle_hole(dim)
        gens.add(PH)
    assert check_bravais_symmetry(neighbors, gens)
    if generators:
        return gens
    else:
        return generate_group(gens)


def bravais_group_2d(neighbors, num_eq, sets_eq, verbose=False):
    s3 = np.sqrt(3)
    gens = set()

    assert len(neighbors) <= 3
    if num_eq == [1, 1] or num_eq == [2]:
        # Sqare or simple rectangular lattice
        name = 'simple rectangular'
        # Mirrors
        Mx = mirror(neighbors[0])
        My = mirror(neighbors[1])
        gens |= {Mx, My}
        if num_eq == [2]:
            # Square lattice, 4-fold rotation
            name = 'square'
            C4 = rotation(1/4)
            gens.add(C4)
    elif num_eq == [1, 2] or num_eq == [3]:
        # Centered rectangular, mirror symmetry
        name = 'centered rectangular'
        vecs = sets_eq[-1][:2]
        Mx = mirror(vecs[0] + vecs[1])
        My = mirror(vecs[0] - vecs[1])
        gens.add(Mx)
        gens.add(My)
        if num_eq == [3]:
            name = 'hexagonal'
            # Hexagonal lattice, 6-fold rotation
            C6 = rotation(1/6)
            gens.add(C6)
    else:
        name = 'oblique'

    if verbose:
        print(name)
    return gens


def bravais_group_3d(neighbors, num_eq, sets_eq, verbose=False):
    assert len(neighbors) <= 7
    gens = set()

    if num_eq == [3]:
        # Primitive cubic, 3 4-fold axes
        name = 'primitive cubic'
        C4s = {rotation(1/4, n) for n in neighbors}
        gens |= C4s
    elif num_eq == [1, 2]:
        # Primitive tetragonal, find 4-fold axis
        name = 'primitive tetragonal'
        C4 = rotation(1/4, np.cross(*sets_eq[1]))
        gens.add(C4)
        C2s = {rotation(1/2, axis) for axis in sets_eq[1]}
        gens |= C2s
    elif num_eq == [1, 1, 1]:
        # Primitive orthorhombic
        name = 'primitive orthorhombic'
        C2s = {rotation(1/2, n) for n in neighbors}
        gens |= C2s
    elif num_eq == [1, 3]:
        # Hexagonal
        name = 'hexagonal'
        # lone one for C6 axis
        C6 = rotation(1/6, sets_eq[0][0])
        # one of the triple for C2 axis
        C2 = rotation(1/2, sets_eq[1][0])
        gens |= {C6, C2}
    elif num_eq == [1, 1, 2]:
        # Base centered orthorhombic
        name = 'base centered orthorhombic'
        vec011, vec01m1 = sets_eq[-1]
        C2x = rotation(1/2, vec011 + vec01m1)
        C2y = rotation(1/2, vec011 - vec01m1)
        C2z = rotation(1/2, np.cross(vec011, vec01m1))
        gens |= {C2x, C2y, C2z}
    elif num_eq == [1, 1, 1, 1]:
        # Primitive Monoclinic
        name = 'primitive monoclinic'
        axis, = pick_perp(neighbors, 3)
        C2 = rotation(1/2, axis)
        gens.add(C2)
    elif num_eq == [3, 4]:
        # Body centered cubic
        name = 'body centered cubic'
        C4s = {rotation(1/4, n) for n in sets_eq[0]}
        gens |= C4s
    elif num_eq == [1, 2, 4] or num_eq == [2, 4]:
        # Body centered tetragonal
        name = 'body centered tetragonal'
        axes = (sets_eq[0] if len(num_eq) == 2 else sets_eq[1])
        C2s = {rotation(1/2, n) for n in axes}
        C4 = rotation(1/4, np.cross(*axes))
        gens |= C2s
        gens.add(C4)
    elif num_eq == [1, 1, 1, 4] or num_eq == [1, 1, 4]:
        # Body centered orthorhombic
        name = 'body centered orthorhombic'
        axes = [vec[0] for num, vec in zip(num_eq, sets_eq) if num == 1]
        C2s = {rotation(1/2, n) for n in axes}
        C2s.add(rotation(1/2, np.cross(axes[0], axes[1])))
        gens |= C2s
    elif num_eq == [6]:
        # FCC
        name = 'face centered cubic'
        # pick an orthogonal pair to define C4 axes
        vec110 = neighbors[0]
        vec1m10, = pick_perp(neighbors, 1, [vec110])
        C4x = rotation(1/4, vec110 + vec1m10)
        C4y = rotation(1/4, vec110 - vec1m10)
        C4z = rotation(1/4, np.cross(vec110, vec1m10))
        gens |= {C4x, C4y, C4z}
    elif num_eq == [1, 3, 3] or num_eq == [3, 3]:
        # Rhombohedral with or without face toward the 3-fold axis
        name = 'rhombohedral'
        C3 = threefold_axis(sets_eq[-1], neighbors)
        assert C3 is not None, sets_eq
        C2 = twofold_axis(sets_eq[-1], neighbors)
        assert C2 is not None, sets_eq
        gens |= {C3, C2}
    elif num_eq == [1, 2, 2, 2]:
        # Face centered orthorhombic
        name = 'face centered orthorhombic'
        # One cubic vector has to be there
        vec100 = sets_eq[0][0]
        C2x = rotation(1/2, vec100)
        # Pick the pair orthogonal to it
        vec011, vec01m1 = pick_perp(neighbors, 1, [vec100])
        C2y = rotation(1/2, vec011 + vec01m1)
        C2z = rotation(1/2, vec011 - vec01m1)
        gens |= {C2x, C2y, C2z}
    elif num_eq[-1] == num_eq[-2] == 2:
        # Base centered monoclinic
        name = 'base centered monoclinic'
        # some combination of the equal pairs has to be 2-fold axis
        for vecs in sets_eq[-2:]:
            C2 = twofold_axis(vecs, neighbors)
            assert C2 is not None
            gens.add(C2)
    else:
        assert max(num_eq) == 1
        name = 'triclinic'

    if verbose:
        print(name)
    return gens


def equals(vectors):
    # group equivalent vectors based on length and angles
    one = kwant_continuum.sympify('1')
    sets = dict()
    # Take abs because every vector has opposite pair
    overlaps = np.abs(vectors @ vectors.T)
    angles = np.outer(np.diag(overlaps), np.diag(overlaps))**(-1/2) * overlaps
    for i, vector in enumerate(vectors):
        length = np.array([overlaps[i, i]])
        # Symmetry equivalent vectors must have the same signature
        signature = np.concatenate([length, sorted(overlaps[i]), sorted(angles[i])])
        key = BlochCoeff(signature, one)
        if key in sets:
            sets[key].append(vector)
        else:
            sets[key] = [vector]
    numbers, sets = zip(*((len(set), np.array(set)) for set in sets.values()))
    order = np.argsort(numbers)
    numbers = np.array(numbers)
    sets = np.array(sets)
    return list(numbers[order]), list(sets[order])


def pick_perp(vectors, n, other_vectors=None):
    # Pick vectors that are orthogonal to at least n other vectors
    other_vectors = np.array(vectors if other_vectors is None else other_vectors)
    perp = []
    non_perp = []
    for v in vectors:
        if np.sum(np.isclose(v @ other_vectors.T, 0)) >= n:
            perp.append(v)
    return perp


def threefold_axis(vectors, neighbors):
    # Find threefold axis that leaves three vectors invariant
    assert len(vectors) == 3
    overlaps = vectors @ vectors.T
    prop, norm = prop_to_id(np.diag(np.diag(overlaps)))
    if not prop:
        return None
    overlaps = np.abs(overlaps) / norm
    if np.allclose([overlaps[0, 1], overlaps[0, 2], overlaps[1, 2]], 1/2):
        # coplanar vectors, may be 6-fold
        axis = np.cross(vectors[0], vectors[1])
        C3 = rotation(1/3, axis)
        C6 = rotation(1/6, axis)
        if check_bravais_symmetry(neighbors, {C6}):
            return C6
        elif check_bravais_symmetry(neighbors, {C3}):
            return C3
        else:
            return None
    for signs in it.product([1, -1], repeat=3):
        axis = signs @ vectors
        overlaps = axis @ vectors.T
        C3 = rotation(1/3, axis)
        prop, _ = prop_to_id(np.diag(np.abs(overlaps)))
        if prop and check_bravais_symmetry(neighbors, {C3}):
            return C3
    else:
        return None


def twofold_axis(vectors, neighbors):
    # Find twofold axis from vectors that leaves neighbors invariant
    for sign, (vec1, vec2) in it.product([1, -1], it.combinations(vectors, 2)):
        axis = vec1 + sign * vec2
        C2 = rotation(1/2, axis)
        if check_bravais_symmetry(neighbors, {C2}):
            return C2
    else:
        return None


def check_bravais_symmetry(neighbors, group):
    one = kwant_continuum.sympify('1')
    neighbors = np.vstack([neighbors, -neighbors])
    neighbors = {BlochCoeff(vec, one) for vec in neighbors}
    for g in group:
        r_neighbors = {BlochCoeff(g.R @ hop, coeff) for (hop, coeff) in neighbors}
        if not neighbors == r_neighbors:
            return False
    else:
        return True



from . import groups
from . import linalg
from . import hamiltonian_generator
from . import symmetry_finder
from . import model
from .groups import (
    PointGroupElement, ContinuousGroupGenerator, identity,
    time_reversal, particle_hole, chiral, inversion, rotation, mirror)
from .model import Model, BlochModel
from .hamiltonian_generator import continuum_hamiltonian, continuum_pairing, display_family, \
								   bloch_family
from .symmetry_finder import symmetries, discrete_symmetries, conserved_quantities, \
                             continuous_symmetries, bravais_point_group
from .kwant_continuum import sympify

from ._version import __version__



from collections import OrderedDict
import numpy as np
import sympy
import itertools as it
import scipy.linalg as la
from copy import deepcopy
import tinyarray as ta

from .linalg import matrix_basis, nullspace, sparse_basis, family_to_vectors, rref, allclose
from .model import Model, BlochModel, _commutative_momenta, e, I
from .groups import PointGroupElement, ContinuousGroupGenerator, generate_group
from . import kwant_continuum


def continuum_hamiltonian(symmetries, dim, total_power, all_powers=None,
                          momenta=_commutative_momenta, sparse_linalg=False,
                          prettify=False, num_digits=10):
    """Generate a family of continuum Hamiltonians that satisfy symmetry constraints.

    Parameters:
    -----------------
    symmetries: iterable of PointGroupElement objects.
        An iterable of PointGroupElement objects, each describing a symmetry
        that the family should possess.
    dim: integer
        The number of spatial dimensions along which the Hamiltonian family is
        translationally invariant. Only the first `dim` entries in `all_powers` and
        `momenta` are used.
    total_power: integer or list of integers
        Allowed total powers of the momentum variables in the continuum Hamiltonian.
        If an integer is given, all powers below it are included as well.
    all_powers: list of integer or list of list of integers
        Allowed powers of the momentum variables in the continuum Hamiltonian listed
        for each spatial direction. If an integer is given, all powers below it are
        included as well. If a list of integers is given, only these powers are used.
    momenta : list of int or list of Sympy objects
        Indices of momenta from ['k_x', 'k_y', 'k_z'] or a list of names for the
        momentum variables. Default is ['k_x', 'k_y', 'k_z'].
    sparse_linalg : bool
        Whether to use sparse linear algebra. Using sparse solver can result in
        performance increase for large, highly constrained families,
        but not as well tested as the default dense version.
    prettify: boolean, default False
        Whether to make the basis pretty by rounding and basis change. For details
        see docstring of `make_basis_pretty`. May be numerically unstable.
    num_digits: integer, default 10
        Number of significant digits to which the basis is rounded using prettify.
        This argument is ignored if prettify = False.

    Returns:
    ---------------
    family: list
        A list of Model objects representing the family that
        satisfies the symmetries specified. Each Model object satisfies
        all the symmetries by construction.
    """

    if type(total_power) is int:
        max_power = total_power
        total_power = range(max_power + 1)

    # Generate a Hamiltonian family
    N = list(symmetries)[0].U.shape[0] # Dimension of Hamiltonian matrix
    momenta = momenta[:dim]
    # Symmetries do not mix the total degree of momenta. We can thus work separately at each
    # fixed degree.
    family = []
    for degree in total_power:
        # Make all momentum variables given the constraints on dimensions and degrees in the family
        momentum_variables = continuum_variables(dim, degree, all_powers=all_powers, momenta=momenta)
        degree_family = [Model({momentum_variable: matrix}, momenta=momenta)
                             for momentum_variable, matrix
                             in it.product(momentum_variables, matrix_basis(N))]
        degree_family = constrain_family(symmetries, degree_family, sparse_linalg=sparse_linalg)
        if prettify:
            family_size = len(degree_family)
            degree_family = make_basis_pretty(degree_family, num_digits=num_digits)
            assert family_size == len(degree_family), 'make_basis_pretty reduced the size of the family, \
                                                possibly due to numerical instability'
        family += degree_family
    return family


def continuum_pairing(symmetries, dim, total_power, all_powers=None, momenta=_commutative_momenta,
                      phases=None, ph_square=1, sparse_linalg=False,
                      prettify=False, num_digits=10):
    """Generate a family of continuum superconducting pairing functions that satisfy
    symmetry constraints.

    The specified symmetry operators should act on the normal state Hamiltonian, not
    in particle-hole space.

    Parameters:
    -----------------
    symmetries: iterable of PointGroupElement objects.
        An iterable of PointGroupElement objects, each describing a symmetry
        that the family should possess.
    dim: integer
        The number of spatial dimensions along which the Hamiltonian family is
        translationally invariant. Only the first `dim` entries in `all_powers` and
        `momenta` are used.
    total_power: integer or list of integers
        Allowed total powers of the momentum variables in the continuum Hamiltonian.
        If an integer is given, all powers below it are included as well.
    all_powers: list of integer or list of list of integers
        Allowed powers of the momentum variables in the continuum Hamiltonian listed
        for each spatial direction. If an integer is given, all powers below it are
        included as well. If a list of integers is given, only these powers are used.
    momenta: list of int or list of Sympy objects
        Indices of momenta from ['k_x', 'k_y', 'k_z'] or a list of names for the
        momentum variables. Default is ['k_x', 'k_y', 'k_z'].
    phases: iterable of numbers
        Phase factors to multiply the hole block of the symmetry operators in
        particle-hole space. By default, all phase factors are 1.
    ph_square: integer, either 1 or -1.
        Specifies whether the particle-hole operator squares to +1 or -1.
    sparse_linalg : bool
        Whether to use sparse linear algebra. Using sparse solver can result in
        performance increase for large, highly constrained families,
        but not as well tested as the default dense version.
    prettify: boolean, default False
        Whether to make the basis pretty by rounding and basis change. For details
        see docstring of `make_basis_pretty`. May be numerically unstable.
    num_digits: integer, default 10
        Number of significant digits to which the basis is rounded using prettify.
        This argument is ignored if prettify = False.
    Returns:
    ---------------
    family: list
        A list of Model objects representing the family that
        satisfies the symmetries specified. Each Model object satisfies
        all the symmetries by construction.
    """
    if type(total_power) is int:
        max_power = total_power
        total_power = range(max_power + 1)

    if not ph_square in (-1, 1):
        raise ValueError('Particle-hole operator must square to +1 or -1.')
    if phases is None:
        phases = np.ones(len(symmetries))
    symmetries = deepcopy(symmetries)
    N = symmetries[0].U.shape[0]
    # Attach phases to symmetry operators and extend to BdG space
    for sym, phase in zip(symmetries, phases):
        if isinstance(sym, PointGroupElement):
            sym.U = la.block_diag(sym.U, phase*sym.U.conj())
        if isinstance(sym, ContinuousGroupGenerator):
            sym.U = la.block_diag(sym.U, -sym.U.conj() + phase*np.eye(N))
    # Build ph operator
    ph = np.array([[0, 1], [ph_square, 0]])
    ph = PointGroupElement(np.eye(dim), True, True, np.kron(ph, np.eye(N)))
    symmetries.append(ph)
    momenta = momenta[:dim]
    momentum_variables = continuum_variables(dim, total_power, all_powers=all_powers, momenta=momenta)
    # matrix basis for all matrices in off-diagonal blocks
    b0 = np.zeros((N, N))
    mbasis = [np.block([[b0, m], [m.T.conj(), b0]]) for m in matrix_basis(N)]
    mbasis.extend([np.block([[b0, 1j*m], [-1j*m.T.conj(), b0]]) for m in matrix_basis(N)])

    # Symmetries do not mix the total degree of momenta. We can thus work separately at each
    # fixed degree.
    family = []
    for degree in total_power:
        # Make all momentum variables given the constraints on dimensions and degrees in the family
        momentum_variables = continuum_variables(dim, degree, all_powers=all_powers, momenta=momenta)
        degree_family = [Model({momentum_variable: matrix}, momenta=momenta)
                             for momentum_variable, matrix
                             in it.product(momentum_variables, mbasis)]
        degree_family = constrain_family(symmetries, degree_family, sparse_linalg=sparse_linalg)
        if prettify:
            family_size = len(degree_family)
            degree_family = make_basis_pretty(degree_family, num_digits=num_digits)
            assert family_size == len(degree_family), 'make_basis_pretty reduced the size of the family, \
                                                possibly due to numerical instability'
        family += degree_family

    # Cast the pairing terms into new Model objects, to ensure that each object has the correct
    # shape.
    family = [Model({term: matrix[:N, N:] for term, matrix in monomial.items()}) for monomial in
              family]
    return [mon for mon in family if len(mon)]


def continuum_variables(dim, total_power, all_powers=None, momenta=_commutative_momenta):
    """Make a list of all linearly independent combinations of momentum
    variables with given total power.

    Parameters
    ----------------
    dim: integer
        The number of spatial dimensions along which the Hamiltonian family is
        translationally invariant. Only the first `dim` entries in `all_powers` and
        `momenta` are used.
    total_power: integer
        Allowed total power of the momentum variables in the continuum Hamiltonian.
    all_powers: list of integer or list of list of integers
        Allowed powers of the momentum variables in the continuum Hamiltonian listed
        for each spatial direction. If an integer is given, all powers below it are
        included as well. If a list of integers is given, only these powers are used.
    momenta : list of int or list of Sympy objects
        Indices of momenta from ['k_x', 'k_y', 'k_z'] or a list of names for the
        momentum variables. Default is ['k_x', 'k_y', 'k_z'].

    Returns
    ---------------
    A list of Sympy objects, representing the momentum variables that enter the Hamiltonian.
    """

    if all_powers is None:
        all_powers = [total_power] * dim

    if len(all_powers) < dim or len(momenta) < dim:
        raise ValueError('`all_powers` and `momenta` must be at least `dim` long.')
    # Only keep the first dim entries
    momenta = momenta[:dim]
    all_powers = all_powers[:dim]

    for i, power in enumerate(all_powers):
        if type(power) is int:
            all_powers[i] = range(power + 1)

    if dim == 0:
        return [kwant_continuum.sympify(1)]

    if all([type(i) is int for i in momenta]):
        momenta = [_commutative_momenta[i] for i in momenta]
    else:
        momenta = [kwant_continuum.make_commutative(k, k)
                        for k in momenta]

    # Generate powers for all terms
    powers = [p for p in it.product(*all_powers) if sum(p) == total_power]
    momentum_variables = [sympy.Mul(*[sympy.Pow(k, power) for k, power in zip(momenta, p)])
                          for p in powers]
    return momentum_variables


def round_family(family, num_digits=3):
    """Round the matrix coefficients of a family to specified significant digits.

    Parameters
    -----------
    family: iterable of Model objects that represents
        a family.
    num_digits: integer
        Number if significant digits to which the matrix coefficients are rounded.

    Returns
    ----------
    A list of Model objects representing the family, with
    the matrix coefficients rounded to num_digits significant digits.

    """
    return [member.around(num_digits) for member in family]


def hamiltonian_from_family(family, coeffs=None, nsimplify=True, tosympy=True):
    """Form a Hamiltonian from a Hamiltonian family by taking a linear combination
    of its elements.

    Parameters
    ----------
    family: iterable of Model or BlochModel objects
        List of terms in the Hamiltonian family.
    coeffs: list of sympy objects, optional
        Coefficients used to form the linear combination of
        terms in the family. Element n of coeffs multiplies
        member n of family. The default choice of the coefficients
        is c_n.
    nsimplify: bool
        Whether to use sympy.nsimplify on the output or not, which
        attempts to replace floating point numbers with simpler expressions,
        e.g. fractions.
    tosympy: bool
        Whether to convert the Hamiltonian to a sympy expression.
        If False, a Model or BlochModel object is returned instead,
        depending on the type of the Hamiltonian family.

    Returns
    -------
    ham: sympy.Matrix or Model/BlochModel object.
        The Hamiltonian, i.e. the linear combination of entries in family.

    """
    if coeffs is None:
        coeffs = list(sympy.symbols('c0:%d'%len(family), real=True))
    else:
        assert len(coeffs) == len(family), 'Length of family and coeffs do not match.'
    ham = sum(c * term for c, term in zip(coeffs, family))
    if tosympy:
        return ham.tosympy(nsimplify=nsimplify)
    else:
        return ham


def display_family(family, summed=False, coeffs=None, nsimplify=True):
    """Helper function to display a Hamiltonian family.
    Supports LaTeX display through Sympy in a Jupyter notebook, which may be enabled
    by running sympy.init_printing(print_builtin=True).

    Parameters
    -----------
    family: iterable of Model or BlochModel objects
        List of terms in a Hamiltonian family.
    summed: boolean
        Whether to display the Hamiltonian family by individual member (False),
        or as a sum over all members with expansion coefficients (True).
    coeffs: list of sympy objects, optional
        Coefficients used when combining terms in the family if summed is True.
    nsimplify: boolean
        Whether to use sympy.nsimplify on the output or not, which attempts to replace
        floating point numbers with simpler expressions, e.g. fractions.
    """

    if not summed:
        # print each member in the family separately
        for term in family:
            sterm = term.tosympy(nsimplify=nsimplify)
            display(sterm)
    else:
        # sum the family members multiplied by expansion coefficients
        display(hamiltonian_from_family(family, coeffs=coeffs,
                                        nsimplify=nsimplify))

def check_symmetry(family, symmetries, num_digits=None):
    """Check that a family satisfies symmetries. A symmetry is satisfied if all members of
    the family satisfy it.

    If the input family was rounded before hand, it is necessary to use
    specify the number of significant digits using num_digits, otherwise
    this check might fail.

    Parameters:
    ------------
    family: iterable of Model or BlochModel objects representing
        a family.
    symmetries: iterable representing the symmetries to check.
        If the family is a Hamiltonian family, symmetries is an iterable
        of PointGroupElement objects representing the symmetries
        to check.
    num_digits: integer
        In the case that the input family has been rounded, num_digits
        should be the number of significant digits to which the family
        was rounded.
    """

    for symmetry in symmetries:
        # Iterate over all members of the family
        for member in family:
            if isinstance(symmetry, PointGroupElement):
                if num_digits is None:
                    assert symmetry.apply(member) == member
                else:
                    assert symmetry.apply(member).around(num_digits) == member.around(num_digits)
            elif isinstance(symmetry, ContinuousGroupGenerator):
                # Continous symmetry if applying it returns zero
                assert symmetry.apply(member) == {}


def constrain_family(symmetries, family, sparse_linalg=False):
    """Apply symmetry constraints to a family.

    Parameters
    -----------
    symmetries: iterable of PointGroupElement objects, representing the symmetries
        that are used to constrain the Hamiltonian family.
    family: iterable of Model or BlochModel objects, representing the Hamiltonian
        family to which the symmetry constraints are applied.
    sparse_linalg : bool
        Whether to use sparse linear algebra. Using sparse solver can result in
        performance increase for large, highly constrained families,
        but not as well tested as the default dense version.

    Returns
    ----------
    family: iterable of Model or BlochModel objects, that represents the
        family with the symmetry constraints applied. """

    if not family:
        return family
    # Fix ordering
    family = list(family)
    symmetries = list(symmetries)
    # Check compatibility of family members and symmetries
    shape = family[0].shape
    momenta = family[0].momenta
    for term in family:
        assert term.shape == shape
        assert term.momenta == momenta
    for symmetry in symmetries:
        assert symmetry.U.shape == shape
        if symmetry.R is not None:
            assert symmetry.R.shape[0] == len(momenta)

    # Need all the linearly independent variables before and after
    # rotation to make the matrix of linear constraints.
    rotated_families = [[symmetry.apply(monomial) for monomial in family]
                                                  for symmetry in symmetries]

    # Get all variables and fix ordering
    all_variables = set()
    for member in it.chain(*rotated_families):
        all_variables |= member.keys()
    all_variables = list(all_variables)

    # Generate the matrix of symmetry constraints.
    constraint_matrices = []
    for i, symmetry in enumerate(symmetries):
        # In block space, each row is the constraint that the matrix coefficient to a linearly independent
        # monomial must vanish. The column index runs over expansion coefficients multiplying different
        # models.
        rotated_family = rotated_families[i]
        constraint_matrix = family_to_vectors(rotated_family, all_keys=all_variables).T
        if isinstance(symmetry, PointGroupElement):
            # Only need to subtract untransformed part for discrete symmetries,
            # continuous symmetry applies the infinitesimal generator and
            # the result should vanish.
            constraint_matrix -= family_to_vectors(family, all_keys=all_variables).T
        constraint_matrices.append(constraint_matrix)

    constraint_matrix = np.vstack(constraint_matrices)
    # If it is empty, there are no constraints
    if not np.any(constraint_matrix):
        return family
    # ROWS of this matrix are the basis vectors
    null_basis = nullspace(constraint_matrix, sparse=sparse_linalg)

    # We return a list of dictionary-like Model objects.
    # Each Model object represents one term in the Hamiltonian family,
    # where keys are the variables and values the matrix coefficients multiplying each variable.
    Hamiltonian_family = []
    for basis_vector in null_basis:
        family_member = sum([val * family[i] for i, val in enumerate(basis_vector)])
        # Eliminate entries that vanish
        if family_member:
            Hamiltonian_family.append(family_member)
    return Hamiltonian_family


def make_basis_pretty(family, num_digits=2):
    """Attempt to make a family more legible by reducing the
    number of nonzero entries in the matrix coefficients.

    Parameters
    -----------
    family: iterable of Model or BlochModel objects representing
        a family.
    num_digits: positive integer
        Number of significant digits that matrix coefficients are rounded to.

    This attempts to make the family more legible by prettifying a matrix,
    which is done by bringing it to reduced row-echelon form. This
    procedure may be numerically unstable, so this function should be used
    with caution. """

    # Return empty family for empty family
    if not family:
        return family
    # convert family to a set of row vectors
    basis_vectors = family_to_vectors(family)
    # Find the transformation that brings it to rref form
    _, S = rref(basis_vectors, return_S=True, rtol=10**(-num_digits))
    # Transform the model by S
    rfamily = []
    for vec in S:
        term = sum([val * family[i] for i, val in enumerate(vec)])
        # Eliminate entries that vanish
        if term:
            rfamily.append(term)
    return round_family(rfamily, num_digits)


def remove_duplicates(family, tol=1e-8):
    """Remove linearly dependent terms in Hamiltonian family using SVD.

    Parameters
    -----------
    family: iterable of Model or BlochModel objects representing
        a family.
    tol: float
        tolerance used in SVD when finding the span.

    Returns
    -------
    rfamily: list of Model or BlochModel objects representing
        the family with only linearly independent terms.
    """
    if not family:
        return family
    # Convert to vectors
    basis_vectors = family_to_vectors(family)
    # Find the linearly independent vectors
    _, basis_vectors = nullspace(basis_vectors.T, atol=tol, return_complement=True)
    rfamily = []
    for vec in basis_vectors:
        rfamily.append(sum([family[i] * c for i, c in enumerate(vec)]))
    return rfamily


def subtract_family(family1, family2, tol=1e-8, prettify=False):
    """Remove the linear span of family2 from the span of family1 using SVD.
    family2 must be a span of terms that are either inside the span of family1
    or orthogonal to it. This guarantees that projecting out family2 from family1
    results in a subfamily of family1.

    Parameters
    -----------
    family1, family2: iterable of Model or BlochModel objects
        Hamiltonian families.
    tol: float
        tolerance used in SVD when finding the span.

    Returns
    -------
    rfamily: list of Model or BlochModel objects representing
        family1 with the span of family2 removed.
    """
    if not family1 or not family2:
        return family1
    # Convert to vectors
    all_keys = set.union(*[set(term.keys()) for term in family1])
    all_keys |= set.union(*[set(term.keys()) for term in family2])
    all_keys = list(all_keys)
    basis1 = family_to_vectors(family1, all_keys=all_keys)
    basis2 = family_to_vectors(family2, all_keys=all_keys)
    # get the orthonormal basis for the span of basis2
    _, basis2 = nullspace(basis2, atol=tol, return_complement=True)
    # project out components in the span of basis2 from basis1
    projected_basis1 = basis1 - (basis1.dot(basis2.T.conj())).dot(basis2)
    # Check that projected_basis1 is a subspace of basis1.
    _, ort_basis1 = nullspace(basis1, atol=tol, return_complement=True)
    if not allclose((projected_basis1.dot(ort_basis1.T.conj())).dot(ort_basis1), projected_basis1, atol=tol):
        raise ValueError('Projecting onto the complement of family2 did not result in a subspace of family1')
    # Find the coefficients of linearly independent vectors
    _, projected_coeffs1 = nullspace(projected_basis1.T, atol=tol, return_complement=True)
    rfamily = []
    for vec in projected_coeffs1:
        rfamily.append(sum([family1[i] * c for i, c in enumerate(vec)]))
    if prettify:
        rfamily = make_basis_pretty(rfamily, num_digits=int(-np.log10(tol)))
    return rfamily


def symmetrize_monomial(monomial, symmetries):
    """Symmetrize monomial by averaging over all symmetry images under symmetries.

    Parameters:
    -----------
    monomial : Model or BlochModel object
        Hamiltonian term to be symmetrized
    symmetries : iterable of PointGroupElement objects
        Symmetries to use for symmetrization. `symmetries` must form a closed group.

    Returns:
    --------
    Model or BlochModel object
        Symmetrized term.
    """
    return sum([sym.apply(monomial) for sym in symmetries]) * (1/len(symmetries))


def bloch_family(hopping_vectors, symmetries, norbs, onsites=True,
                 symmetrize=True, prettify=True, num_digits=10,
                 bloch_model=False):
    """Generate a family of symmetric Bloch-Hamiltonians.

    Parameters:
    -----------
    hopping_vectors : list of tuples (a, b, vec)
        `a` and `b` are identifiers for the different sites (e.g. strings) of
        the unit cell, `vec` is the real space hopping vector. Vec may contain
        contain integers, sympy symbols, or floating point numbers.
    symmetries : list of PointGroupElement or ContinuousGroupGenerator
        Generators of the symmetry group. ContinuousGroupGenerators can only
        have onsite action as a lattice system cannot have continuous rotation
        invariance. It is assumed that the block structure of the unitary action
        is consistent with norbs, as returned by `symmetry_from_permutation`.
    norbs : OrderedDict : {site : norbs_site} or tuple of tuples ((site, norbs_site), )
        sites are ordered in the order specified, with blocks of size norbs_site
        corresponding to each site.
    onsites : bool
        Whether to include on-site terms consistent with the symmetry.
    symmetrize : bool
        Whether to use the symmetrization strategy. This does not require
        a full set of hoppings to start, all symmetry related hoppings
        are generated. Otherwise the constraining strategy is used, this does
        not generate any new hoppings beyond the ones specified and constrains
        coefficients to enforce symmetry.
    prettify: bool
        Whether to prettify the result. This step may be numerically unstable.
    num_digits: int, default 10
        Number of significant digits kept when prettifying.
    bloch_model: bool, default False
        Determines the return format of this function. If set to False, returns
        a list of Model objects. If True, returns a list of BlochModel objects.
        BlochModel objects are more suitable than Model objects if the hopping
        vectors include floating point numbers.

    Returns:
    --------
    family: list of Model or BlochModel objects
        A list of Model or BlochModel objects representing the family that
        satisfies the symmetries specified. Each object satisfies
        all the symmetries by construction.

    Notes:
    ------
    There is no need to specify the translation vectors, all necessary information
    is encoded in the symmetries and hopping vectors.

    In the generic case the Bloch-Hamiltonian produced is not Brillouin-zone periodic,
    instead it acquires a unitary transformation `W_G = delta_{ab} exp(i G (r_a - r_b))`
    where `G` is a reciprocal lattice vector, `a` and `b` are sites and `r_a` is the
    real space position of site `a`. If the lattice is primitive (there is only one
    site per unit cell), the hopping vectors are also translation vectors and the
    resulting Hamiltonian is BZ periodic.

    If `symmetrize=True`, all onsite unitary symmetries need to be explicitely
    specified as ContinuousGroupGenerators. Onsite PointGroupSymmetries (ones
    with R=identity) are ignored.

    If floating point numbers are used in the argument hopping_vectors, it is
    recommended to have this function return BlochModel objects instead of Model
    objects, by setting the bloch_model flag to True.
    """

    N = 0
    if not any([isinstance(norbs, OrderedDict), isinstance(norbs, list),
                isinstance(norbs, tuple)]):
        raise ValueError('norbs must be OrderedDict, tuple, or list.')
    else:
        norbs = OrderedDict(norbs)
    ranges = dict()
    for a, n in norbs.items():
        ranges[a] = slice(N, N + n)
        N += n
    # Separate point group and conserved quantites
    pg = [g for g in symmetries if isinstance(g, PointGroupElement)]
    conserved = [g for g in symmetries if isinstance(g, ContinuousGroupGenerator)]
    if not all([(g.R is None or np.allclose(g.R, np.zeros_like(g.R))) for g in conserved]):
        raise ValueError('Bloch Hamiltonian cannot have continuous rotation symmetry.')
    if (not bloch_model) and any(isinstance(g.R, ta.ndarray_float) for g in pg):
        raise ValueError('Cannot generate Bloch Hamiltonian in Model format using '
                         'floating point rotation matrices. To avoid this error, use '
                         'only PointGroupElements that are defined with exact sympy '
                         'rotation matrices or use `bloch_model=True`.')

    # Check dimensionality
    dim = len(hopping_vectors[0][-1])
    assert all([len(hop[-1]) == dim for hop in hopping_vectors])

    # Add zero hoppings for onsites
    if onsites:
        hopping_vectors = deepcopy(hopping_vectors)
        hopping_vectors += [(a, a, [0] * dim) for a in norbs]

    family = []
    for a, b, vec in hopping_vectors:
        n, m = norbs[a], norbs[b]
        block_basis = np.eye(n*m, n*m).reshape((n*m, n, m))
        block_basis = np.concatenate((block_basis, 1j*block_basis))
        # Hopping direction in real space
        # Dot product with momentum vector
        phase = sum([coordinate * momentum for coordinate, momentum in
                     zip(vec, _commutative_momenta[:dim])])
        factor = e**(I*phase)
        hopfamily = []
        for mat in block_basis:
            matrix = np.zeros((N, N), dtype=complex)
            matrix[ranges[a], ranges[b]] = mat
            term = Model({factor: matrix}, momenta=('k_x', 'k_y', 'k_z')[:dim])
            term = term + term.T().conj()
            hopfamily.append(term)
        # If there are conserved quantities, constrain the hopping, it is assumed that
        # conserved quantities do not mix different sites
        if conserved:
            hopfamily = constrain_family(conserved, hopfamily)
        family.extend(hopfamily)
    # Use BlochModel objects instead of Model.
    if bloch_model:
        family = [BlochModel(member, momenta=member.momenta) for
                  member in family]
    if symmetrize:
        # Make sure that group is generated while keeping track of unitary part.
        for g in pg:
            g._strict_eq = True
        pg = generate_group(set(pg))
        # Symmetrize every term and remove linearly dependent or zero ones
        family2 = []
        for term in family:
            term = symmetrize_monomial(term, pg).around(decimals=num_digits)
            if not term == {}:
                family2.append(term)
        family = remove_duplicates(family2, tol=10**(-num_digits))
    else:
        # Constrain the terms by symmetry
        family = constrain_family(pg, remove_duplicates(family))
    if prettify:
        family = make_basis_pretty(family, num_digits=num_digits)
    return family



import itertools as it
import pytest
import sympy

from .. import kwant_rmt
from ..symmetry_finder import *
from ..symmetry_finder import _reduced_model, _reduce_hamiltonian, bravais_point_group
from ..linalg import *
from .. import kwant_continuum

sigma = np.array([[[1, 0], [0, 1]], [[0, 1], [ 1, 0]], [[0, -1j], [1j, 0]], [[1, 0], [0, -1]]])

# Stuff for testing
def productrep(n, gens):
    # Make product of n copies of LA rep defined by gens
    genn = gens
    for _ in range(n-1):
        genn = np.array([(np.kron(genn[i],np.eye(len(gens[i])))+
                          np.kron(np.eye(len(genn[i])),gens[i])) for i in range(len(gens))])
    return genn


def sumrep(*args):
    # Make direct sum of n LA reps defined by list of LA's args
    genn = np.array([la.block_diag(*[args[j][i] for j in range(len(args))])
                    for i in range(len(args[0]))])
    return genn


def test_cont_finder():
    # Test symmetry adapted basis
    gens = sumrep((*2*[0.5*sigma[[3, 1, 2]]]))
    U2 = kwant_rmt.circular(len(gens[0]))
    gens2 = np.einsum('ij,ajk,kl->ail',(U2.T).conjugate(),gens,U2)
    U = symmetry_adapted_sun(gens, check=True)
    U = np.hstack(U)
    gensr = np.einsum('ij,ajk,kl->ail',(U.T).conjugate(),gens,U)
    U = symmetry_adapted_sun(gens2, check=True)
    U = np.hstack(U)
    gens2r = np.einsum('ij,ajk,kl->ail',(U.T).conjugate(),gens2,U)
    # check reduced generators are identical
    assert np.allclose(gensr, gens2r), 'Reduced generators should be identical'

    # Test solve_mat_eqn
    n = np.random.randint(2, 5)
    h1 = np.random.random((2, n, n))
    h2 = np.random.random((2, n, n))

    H = np.array([la.block_diag(h1[i], h1[i]) for i in range(len(h1))])
    H = H + (np.einsum('ijk->ikj', H)).conjugate()
    assert solve_mat_eqn(H, H, hermitian=True, traceless=True).shape == (3, 2*n, 2*n)
    assert solve_mat_eqn(H, H, hermitian=False, traceless=True).shape == (3, 2*n, 2*n)
    assert solve_mat_eqn(H, H, hermitian=True, traceless=False).shape == (4, 2*n, 2*n)

    H = np.array([la.block_diag(h1[i], h1[i], h1[i]) for i in range(len(h1))])
    H = H + (np.einsum('ijk->ikj', H)).conjugate()
    assert solve_mat_eqn(H, H, hermitian=True, traceless=True).shape == (8, 3*n, 3*n)
    assert solve_mat_eqn(H, H, hermitian=False, traceless=True).shape == (8, 3*n, 3*n)
    assert solve_mat_eqn(H, H, hermitian=True, traceless=False).shape == (9, 3*n, 3*n)

    H = np.array([la.block_diag(h1[i], h2[i]) for i in range(len(h1))])
    H = H + (np.einsum('ijk->ikj', H)).conjugate()
    assert solve_mat_eqn(H, H, hermitian=True, traceless=True).shape == (1, 2*n, 2*n)
    assert solve_mat_eqn(H, H, hermitian=False, traceless=True).shape == (1, 2*n, 2*n)
    assert solve_mat_eqn(H, H, hermitian=True, traceless=False).shape == (2, 2*n, 2*n)

    # Test _reduce_hamiltonian
    Hs = [h1, np.array([la.block_diag(h1[i], h1[i]) for i in range(len(h1))]),
         np.array([la.block_diag(h1[i], h2[i]) for i in range(len(h1))])]
    for H in Hs:
        H = H + (np.einsum('ijk->ikj', H)).conjugate()
        Ps = _reduce_hamiltonian(H)
        # Check it is proportional to the identity in every block
        for P in Ps:
            Hr = mtm(P[0].T.conjugate(), H, P[0])
            for i, j in it.product(range(len(P)), repeat=2):
                if i ==j:
                    assert np.allclose(mtm(P[i].T.conjugate(), H, P[j]), Hr)
                else:
                    assert np.allclose(mtm(P[i].T.conjugate(), H, P[j]), 0)
        # Check that it vanishes between every block
        for P1, P2 in it.combinations(Ps, 2):
            P1, P2 = np.hstack(P1), np.hstack(P2)
            assert np.allclose(mtm(P1.T.conjugate(), H, P2), 0)


def test_disc_finder(verbose = False):
    """Many test cases for finding discrete onsite symmetries in 0 dimensions"""

    # Thorough testing for all symmetry classes with a single block in a random basis
    sym_list = 'A', 'AI', 'AII', 'AIII', 'BDI', 'CII', 'D', 'DIII', 'C', 'CI'
    dim = 8  # Dimension of each block
             # need 8 because must be multiple of 4 for CII and DIII has symmetry with dim<=4
    for sym in sym_list:
        Hs = []
        for _ in range(4):
            h = kwant_rmt.gaussian(dim, sym)
            Hs.append(h)
        Hs = np.array(Hs)
        # Randomize basis
        U = kwant_rmt.circular(dim)
        Hs = np.einsum('ij,ajk,kl->ail',U.T.conj(),Hs,U)
        Hs = Model({kwant_continuum.sympify('a_' + str(i)) : H for i, H in enumerate(Hs)},
                           momenta=[])
        # Find symmetry operators
        sg = {time_reversal(0), particle_hole(0), chiral(0)}
        sg, Ps = discrete_symmetries(Hs, sg)
        if verbose:
            print('sym', sg)
        T, P, C = kwant_rmt.t(sym), kwant_rmt.p(sym), kwant_rmt.c(sym)
        # check that there are no extra symmetries
        assert len(Ps) == 1
        assert [P.shape for P in Ps] == [(1, dim, dim)]
        assert np.sum(np.abs([T, P, C])) + 1 == len(sg), print([(g.conjugate, g.antisymmetry) for g in sg], T, P, C)
        if T:
            # Check that there is TR
            s = (True, False)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S.conj())
            assert np.allclose(S2, T*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            assert g.apply(Hs) == Hs
            for H in Hs.values():
                assert np.allclose(H, S.dot(H.conj()).dot(S.T.conj()))
        if P:
            # Check that there is TR
            s = (True, True)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S.conj())
            assert np.allclose(S2, P*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            for H in Hs.values():
                assert np.allclose(-H, S.dot(H.conj()).dot(S.T.conj()))
        if C:
            # Check that there is TR
            s = (False, True)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S)
            assert np.allclose(S2, C*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            for H in Hs.values():
                assert np.allclose(-H, S.dot(H).dot(S.T.conj()))
    ##############

    # Thorough testing for all symmetry classes with a two different blocks
    # with the same symmetry in a random basis, also tests _reduce_hamiltonian
    sym_list = 'A', 'AI', 'AII', 'AIII', 'BDI', 'CII', 'D', 'C', 'CI'
    dim = 4  # Dimension of each block
             # skip DIII because it has symmetry with dim<=4
    for sym in sym_list:
        Hs = []
        for _ in range(4):
            h1 = kwant_rmt.gaussian(dim, sym)
            h2 = kwant_rmt.gaussian(dim, sym)
            Hs.append(la.block_diag(h1, h2))
        Hs = np.array(Hs)
        # Randomize basis
        U = kwant_rmt.circular(2*dim)
        Hs = np.einsum('ij,ajk,kl->ail',U.T.conj(),Hs,U)
        Hs = Model({kwant_continuum.sympify('a_' + str(i)) : H for i, H in enumerate(Hs)},
                           momenta=[])
        # Find symmetry operators
        sg = {time_reversal(0), particle_hole(0), chiral(0)}
        sg, Ps = discrete_symmetries(Hs, sg)
        if verbose:
            print('sym', sg)
        T, P, C = kwant_rmt.t(sym), kwant_rmt.p(sym), kwant_rmt.c(sym)
        # check that there are exactly two blocks
        assert len(Ps) == 2
        assert [P.shape for P in Ps] == [(1, 2*dim, dim), (1, 2*dim, dim)]
        assert np.sum(np.abs([T, P, C])) + 1 == len(sg), print([(g.conjugate, g.antisymmetry) for g in sg], T, P, C)
        if T:
            # Check that there is TR
            s = (True, False)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S.conj())
            assert np.allclose(S2, T*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            assert g.apply(Hs) == Hs
            for H in Hs.values():
                assert np.allclose(H, S.dot(H.conj()).dot(S.T.conj()))
        if P:
            # Check that there is TR
            s = (True, True)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S.conj())
            assert np.allclose(S2, P*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            for H in Hs.values():
                assert np.allclose(-H, S.dot(H.conj()).dot(S.T.conj()))
        if C:
            # Check that there is TR
            s = (False, True)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S)
            assert np.allclose(S2, C*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            for H in Hs.values():
                assert np.allclose(-H, S.dot(H).dot(S.T.conj()))
    ##############

    # Thorough testing for all symmetry classes with a two identical blocks
    # with the same symmetry in a random basis, also tests _reduce_hamiltonian
    # The square of TR and PH may change, only thest 4th power
    sym_list = 'A', 'AI', 'AII', 'AIII', 'BDI', 'CII', 'D', 'C', 'CI'
    dim = 4  # Dimension of each block
             # skip DIII because it has symmetry with dim<=4
    for sym in sym_list:
        Hs = []
        for _ in range(4):
            h = kwant_rmt.gaussian(dim, sym)
            Hs.append(la.block_diag(h, h))
        Hs = np.array(Hs)
        # Randomize basis
        U = kwant_rmt.circular(2*dim)
        Hs = np.einsum('ij,ajk,kl->ail',U.T.conj(),Hs,U)
        Hs = Model({kwant_continuum.sympify('a_' + str(i)) : H for i, H in enumerate(Hs)},
                           momenta=[])
        # Find symmetry operators
        sg = {time_reversal(0), particle_hole(0), chiral(0)}
        sg, Ps = discrete_symmetries(Hs, sg)
        if verbose:
            print('sym', sg)
        T, P, C = kwant_rmt.t(sym), kwant_rmt.p(sym), kwant_rmt.c(sym)
        # check that there are exactly two blocks
        assert len(Ps) == 1
        assert [P.shape for P in Ps] == [(2, 2*dim, dim)], (T, P, C)
        assert np.sum(np.abs([T, P, C])) + 1 == len(sg), print([(g.conjugate, g.antisymmetry) for g in sg], T, P, C)
        if T:
            # Check that there is TR
            s = (True, False)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S.conj())
            assert np.allclose(S2, T*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            assert g.apply(Hs) == Hs
            for H in Hs.values():
                assert np.allclose(H, S.dot(H.conj()).dot(S.T.conj()))
        if P:
            # Check that there is TR
            s = (True, True)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S.conj())
            assert np.allclose(S2, P*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            for H in Hs.values():
                assert np.allclose(-H, S.dot(H.conj()).dot(S.T.conj()))
        if C:
            # Check that there is TR
            s = (False, True)
            g = [g for g in sg if (g.conjugate, g.antisymmetry) == s]
            assert len(g) == 1
            g = g[0]
            # Check square
            S = g.U
            S2 = S.dot(S)
            assert np.allclose(S2, C*np.eye(S2.shape[0]))
            # Check that it is a symmetry
            for H in Hs.values():
                assert np.allclose(-H, S.dot(H).dot(S.T.conj()))
    ##############

    # Two blocks related by time reversal symmetry
    dim = 4  # Dimension of each block
    sym = 'AII'  # Time-reversal squares to -1
    # Time reversal operator for each block
    t_mat = np.array(kwant_rmt.h_t_matrix[sym])
    t_mat = np.kron(np.identity(dim // len(t_mat)), t_mat)
    # Hamiltonians
    # Consider a family of Hamiltonians with two blocks related by time-reversal
    Hs = []
    for _ in range(4):
        h = kwant_rmt.gaussian(dim) # Random Hamiltonian with no symmetry
        # Blocks related by time-reversal
        Hs.append(la.block_diag(h, t_mat.dot(h.conj()).dot(t_mat.T.conj())))
    # Complete time-reversal operator
    sx = np.array([[0, 1], [1, 0]])
    T = np.kron(sx, t_mat)
    for H in Hs:
        assert np.allclose(H, T.dot(H.conj()).dot(T.T.conj()))
    # Find the symmetry operator
    Hs = Model({kwant_continuum.sympify('a_' + str(i)) : H for i, H in enumerate(Hs)},
                           momenta=[])
    sg = {time_reversal(0), particle_hole(0), chiral(0)}
    sg, Ps = discrete_symmetries(Hs, sg)
    assert [P.shape for P in Ps] == [(1, 2*dim, dim), (1, 2*dim, dim)]
    assert len(sg) == 2
    tr = list(sg)[0]
    TR = tr.U
    # Squares to +1
    assert np.allclose(TR.dot(TR.conj()), np.eye(TR.shape[0]))

    if verbose:
        print('TR**2', TR.dot(TR.conj()))
    assert np.allclose(TR.dot(TR.conj()).dot(TR.dot(TR.conj())), np.eye(TR.shape[0]))
    # Is a symmetry of all Hamiltonians
    for H in Hs:
        assert tr.apply(Hs) == Hs
    ##########################

    # Add a third block, one which has time-reversal symmetry by itself, but is not related to the other two.
    dim = 4  # Dimension of each block
    sym = 'AII'  # Time-reversal squares to -1
    # Time reversal operator for each block
    t_mat = np.array(kwant_rmt.h_t_matrix[sym])
    t_mat = np.kron(np.identity(dim // len(t_mat)), t_mat)

    # Consider a family of Hamiltonians with two blocks related by time-reversal,
    # neither of which has time-reversal, and a third block that has time-reversal.
    Hs = []
    for _ in range(8):
        h = kwant_rmt.gaussian(dim) # Random Hamiltonian with no symmetry
        ht = kwant_rmt.gaussian(dim, sym)
        Hs.append(la.block_diag(ht, h, t_mat.dot(h.conj()).dot(t_mat.T.conj())))
    # Complete time-reversal operator
    sx = np.array([[0, 1], [1, 0]])
    T = la.block_diag(t_mat, np.kron(sx, t_mat))
    for H in Hs:
        assert np.allclose(H, T.dot(H.conj()).dot(T.T.conj()))
    Hs = Model({kwant_continuum.sympify('a_' + str(i)) : H for i, H in enumerate(Hs)},
                           momenta=[])
    sg = {time_reversal(0), particle_hole(0), chiral(0)}
    sg, Ps = discrete_symmetries(Hs, sg)
    assert [P.shape for P in Ps] == [(1, 3*dim, dim), (1, 3*dim, dim), (1, 3*dim, dim)]
    assert len(sg) == 2
    tr = list(sg)[0]
    TR = tr.U
    if verbose:
        print('TR**2', TR.dot(TR.conj()))
    assert np.allclose(TR.dot(TR.conj()).dot(TR.dot(TR.conj())), np.eye(TR.shape[0]))
    # Is a symmetry of all Hamiltonians
    for H in Hs:
        assert tr.apply(Hs) == Hs
    ##########################

    # Two blocks related by time reversal symmetry, with each block having
    # time-reversal symmetry. Here, there are two ways to consruct the
    # time-reversal operator and there is a unitary symmetry relating the blocks
    dim = 4  # Dimension of each block
    sym = 'AII'  # Time-reversal squares to -1
    # Time reversal operator for each block
    t_mat = np.array(kwant_rmt.h_t_matrix[sym])
    t_mat = np.kron(np.identity(dim // len(t_mat)), t_mat)
    # Hamiltonians
    # Consider a family of Hamiltonians with two blocks related by time-reversal
    Hs = []
    for _ in range(8):
        h = kwant_rmt.gaussian(dim, sym) # Random Hamiltonian with time-reversal
        # Blocks related by time-reversal, each with time-reversal
        Hs.append(la.block_diag(h, t_mat.dot(h.conj()).dot(t_mat.T.conj())))
    # Complete time-reversal operator
    sx = np.array([[0, 1], [1, 0]])
    T = np.kron(sx, t_mat)
    for H in Hs:
        assert np.allclose(H, T.dot(H.conj()).dot(T.T.conj()))
    Hs = Model({kwant_continuum.sympify('a_' + str(i)) : H for i, H in enumerate(Hs)},
                           momenta=[])
    sg = {time_reversal(0), particle_hole(0), chiral(0)}
    sg, Ps = discrete_symmetries(Hs, sg)
    assert [P.shape for P in Ps] == [(2, 2*dim, dim)]
    assert len(sg) == 2
    tr = list(sg)[0]
    TR = tr.U
    if verbose:
        print('TR**2', TR.dot(TR.conj()))
    assert (np.allclose(TR.dot(TR.conj()), np.eye(TR.shape[0])) or
            np.allclose(TR.dot(TR.conj()), -np.eye(TR.shape[0])))
    # Is a symmetry of all Hamiltonians
    for H in Hs:
        assert tr.apply(Hs) == Hs
    ############

    # Four blocks. The first two related by time-reversal but neither
    # possessing it themselves. The second two both possess time-reversal
    # and are related by it.
    dim = 4  # Dimension of each block
    sym = 'AII'  # Time-reversal squares to -1
    # Time reversal operator for each block
    t_mat = np.array(kwant_rmt.h_t_matrix[sym])
    t_mat = np.kron(np.identity(dim // len(t_mat)), t_mat)

    # Hamiltonians
    # Consider a family of Hamiltonians with two blocks related by time-reversal
    Hs = []
    for _ in range(8):
        h1 = kwant_rmt.gaussian(dim, sym) # Random Hamiltonian with time-reversal
        h2 = kwant_rmt.gaussian(dim) # Random Hamiltonian with no symmetry
        Hs.append(la.block_diag(h2, t_mat.dot(h2.conj()).dot(t_mat.T.conj()),
                                h1, t_mat.dot(h1.conj()).dot(t_mat.T.conj())))
    Hs = Model({kwant_continuum.sympify('a_' + str(i)) : H for i, H in enumerate(Hs)},
                           momenta=[])
    sg = {time_reversal(0), particle_hole(0), chiral(0)}
    sg, Ps = discrete_symmetries(Hs, sg)
    assert sorted([P.shape for P in Ps]) == sorted([(1, 4*dim, dim), (1, 4*dim, dim), (2, 4*dim, dim)])
    assert len(sg) == 2
    tr = list(sg)[0]
    TR = tr.U
    if verbose:
        print('TR**2', TR.dot(TR.conj()))
    assert np.allclose(TR.dot(TR.conj()).dot(TR.dot(TR.conj())), np.eye(TR.shape[0]))
    # Is a symmetry of all Hamiltonians
    for H in Hs:
        assert tr.apply(Hs) == Hs
    ##########################


def test_simult_diag():
    # Make diagonal unitary and hermitian matrices with degenerate eigenvalues
    Hs = [np.diag(np.exp(1j * np.random.randint(-3, 3, (20)))) for _ in range(3)]
    Hs += [np.diag(np.random.randint(-3, 3, (20))) for _ in range(3)]
    # random unitary rotation
    U = kwant_rmt.circular(len(Hs[0]))
    Hs = U @ Hs @ U.T.conj()
    # undo the rotation with simult_diag
    U = np.hstack(simult_diag(Hs))
    Hsd = (U.T.conj() @ Hs @ U)
    # result is diagonal
    assert np.allclose(np.abs([H - np.diag(H.diagonal()) for H in Hsd]), 0)


def test_continuum():
    # Simple tests for continuum models

    # Cubic point group
    eye = np.array(np.eye(3), int)
    E = PointGroupElement(eye, False, False, None)
    I = PointGroupElement(-eye, False, False, None)
    C4 = PointGroupElement(np.array([[1, 0, 0],
                                    [0, 0, 1],
                                    [0, -1, 0]], int), False, False, None)
    C3 = PointGroupElement(np.array([[0, 0, 1],
                                    [1, 0, 0],
                                    [0, 1, 0]], int), False, False, None)
    TR = time_reversal(realspace_dim=3)
    PH = particle_hole(realspace_dim=3)
    cubic_gens = {I, C4, C3, TR, PH}
    cubic_group = generate_group(cubic_gens)
    assert len(cubic_group) == 192

    # First model
    ham1 = ("hbar^2 / (2 * m) * (k_x**2 + k_y**2 + k_z**2) * eye(2) +" +
        "alpha * sigma_x * k_x + alpha * sigma_y * k_y + alpha * sigma_z * k_z")
    # Convert to standard model form
    H1 = Model(ham1)
    sg, Ps = discrete_symmetries(H1, cubic_group)
    assert [P.shape for P in Ps] == [(1, 2, 2)]
    assert len(sg) == 48
    assert sg == generate_group({C4, C3, TR})

    # Add a degeneracy
    ham2 = "kron(eye(2), " + ham1 + ")"
    # Convert to standard model form
    H2 = Model(ham2)
    sg, Ps = discrete_symmetries(H2, cubic_group)
    assert [P.shape for P in Ps] == [(2, 4, 2)]
    assert len(sg) == 48
    assert sg == generate_group({C4, C3, TR})

    # Add hole degrees of freedom
    ham2 = "kron(sigma_z, " + ham1 + ")"
    # Convert to standard model form
    H3 = Model(ham2)
    sg, Ps = discrete_symmetries(H3, cubic_group)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 96
    assert sg == generate_group({C4, C3, TR, PH})

    # Continuous rotation symmetry
    for H in [H1, H2, H3]:
        assert len(continuous_symmetries(H)) == 3

    # Test sparse
    H3sp = H3.tocsr()
    sg, Ps = discrete_symmetries(H3, cubic_group, sparse_linalg=True)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 96
    assert sg == generate_group({C4, C3, TR, PH})
    assert len(continuous_symmetries(H, sparse_linalg=True)) == 3
    sg, Ps = discrete_symmetries(H3sp, cubic_group, sparse_linalg=True)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 96
    assert sg == generate_group({C4, C3, TR, PH})
    assert len(continuous_symmetries(H, sparse_linalg=True)) == 3
    sg, Ps = discrete_symmetries(H3sp, cubic_group, sparse_linalg=False)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 96
    assert sg == generate_group({C4, C3, TR, PH})
    assert len(continuous_symmetries(H, sparse_linalg=True)) == 3


def test_bloch():
    # Simple tests for Bloch models

    # Hexagonal point group
    eyesym = sympy.ImmutableMatrix(sympy.eye(2))
    Mx = PointGroupElement(sympy.ImmutableMatrix([[-1, 0],
                                                  [0, 1]]),
                            False, False, None)
    C6 = PointGroupElement(sympy.ImmutableMatrix([[sympy.Rational(1, 2), sympy.sqrt(3)/2],
                                                  [-sympy.sqrt(3)/2,       sympy.Rational(1, 2)]]),
                                 False, False, None)
    TR = time_reversal(realspace_dim=2)
    PH = particle_hole(realspace_dim=2)
    gens_hex_2D ={Mx, C6, TR, PH}
    hex_group_2D = generate_group(gens_hex_2D)
    assert len(hex_group_2D) == 48

    # First example
    ham6 = 'm * (cos(k_x) + cos(1/2*k_x + sqrt(3)/2*k_y) + cos(-1/2*k_x + sqrt(3)/2*k_y))'
    H6 = Model(ham6, momenta=['k_x', 'k_y'])
    sg, Ps = discrete_symmetries(H6, hex_group_2D)
    assert [P.shape for P in Ps] == [(1, 1, 1)]
    assert len(sg) == 24
    assert sg == generate_group({Mx, C6, TR})

    # extend model to add SOC
    ham62 = 'eye(2) * (' + ham6 + ') +'
    ham62 += 'alpha * (sin(k_x) * sigma_x + sin(1/2*k_x + sqrt(3)/2*k_y) * (1/2*sigma_x + sqrt(3)/2*sigma_y) +'
    ham62 += 'sin(-1/2*k_x + sqrt(3)/2*k_y) * (-1/2*sigma_x + sqrt(3)/2*sigma_y))'
    H62 = Model(ham62, momenta=['k_x', 'k_y'])
    sg, Ps = discrete_symmetries(H62, hex_group_2D)
    assert [P.shape for P in Ps] == [(1, 2, 2)]
    assert len(sg) == 24
    assert sg == generate_group({Mx, C6, TR})

    # Add degeneracy
    ham63 = 'kron(eye(2), ' + ham62 + ')'
    H63 = Model(ham63, momenta=['k_x', 'k_y'])
    sg, Ps = discrete_symmetries(H63, hex_group_2D)
    assert [P.shape for P in Ps] == [(2, 4, 2)]
    assert len(sg) == 24
    assert sg == generate_group({Mx, C6, TR})

    # Add PH states
    ham64 = 'kron(sigma_z, ' + ham62 + ')'
    H64 = Model(ham64, momenta=['k_x', 'k_y'])
    sg, Ps = discrete_symmetries(H64, hex_group_2D)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 48
    assert sg == generate_group({Mx, C6, TR, PH})

    # Test sparse
    H64 = Model(ham64, momenta=['k_x', 'k_y'])
    sg, Ps = discrete_symmetries(H64, hex_group_2D, sparse_linalg=True)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 48
    assert sg == generate_group({Mx, C6, TR, PH})
    Hcsr = H64.tocsr()
    sg, Ps = discrete_symmetries(Hcsr, hex_group_2D, sparse_linalg=True)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 48
    assert sg == generate_group({Mx, C6, TR, PH})
    sg, Ps = discrete_symmetries(Hcsr, hex_group_2D, sparse_linalg=False)
    assert [P.shape for P in Ps] == [(1, 4, 2), (1, 4, 2)]
    assert len(sg) == 48
    assert sg == generate_group({Mx, C6, TR, PH})


def test_bravais_symmetry():
    # 2D
    # random rotation
    R = kwant_rmt.circular(2, sym='D', rng=1)
    lattices = [
                ([[np.sqrt(3)/2, 1/2], [0, 1]], 12, 'hexagonal'),
                ([[1, 0], [0, 1]], 8, 'square'),
                ([[-1/3, 1], [1/3, 1]], 4, 'centered orthorhombic')
                ]
    for periods, n, name in lattices:
        group = bravais_point_group(periods, tr=False, ph=False)
        assert len(group) == n, (name, periods, group, n)
        group = bravais_point_group(periods @ R, tr=False, ph=False)
        assert len(group) == n, (name, periods, group, n)

    # 3D
    # random rotation
    R = kwant_rmt.circular(3, sym='D', rng=1)
    lattices = [
                (np.eye(3), 48, 'primitive cubic'),
                ([[1, 0, 0], [0, 1, 0], [1/2, 1/2, 1/2]], 48, 'BCC'),
                ([[1, 1, 0], [1, 0, 1], [0, 1, 1]], 48, 'FCC'),
                ([[1, 0, 0], [0, 1, 0], [0, 0, 3]], 16, 'primitive tetragonal'),
                ([[1, 0, 0], [0, 1, 0], [1/2, 1/2, 3]], 16, 'body centered tetragonal'),
                ([[1, 0, 0], [0, 2, 0], [0, 0, 3]], 8, 'primitive orthorhombic'),
                ([[1, 0, 0], [0, 2, 0], [1/2, 1, 3]], 8, 'body centered orthorhombic'),
                ([[10, 3, 0], [10, 0, 4], [0, 3, 4]], 8, 'face centered orthorhombic'),
                ([[1, 3, 0], [1, 0, 4], [0, 3, 4]], 8, 'face centered orthorhombic'),
                ([[1, 1/3, 0], [1, -1/3, 0], [0, 0, 4]], 8, 'base centered orthorhombic'),
                ([[1, 1/3, 0], [1, -1/3, 0], [0, 0, np.sqrt(10)/3]], 8, 'base centered orthorhombic corner case'),
                ([[1, 0, 0], [0, 2, 0], [1/10, 0, 4]], 4, 'primitive monoclinic'),
                ([[1, 0, 0], [0, 1, 0], [0, 1/10, 2]], 4, 'primitive monoclinic'),
                ([[1, 1/3, 0], [1, -1/3, 0], [1/10, 0, 4]], 4, 'base centered monoclinic'),
                ([[3, 0, 1], [3/2, 1, 1/2], [0, 0, 4]], 4, 'base centered monoclinic'),
                ([[3, 0, 1], [3/2, 1/2, 1/2], [0, 0, 4]], 4, 'base centered monoclinic'),
                ([[1, 0, 1/10], [0, 2, 1/10], [0, 0, 3]], 2, 'triclinic'),
                ([[1, 0, 1/5], [-1/2, np.sqrt(3)/2, 1/5], [-1/2, -np.sqrt(3)/2, 1/5]], 12, 'rhombohedral'),
                ([[1, 0, 5], [-1/2, np.sqrt(3)/2, 5], [-1/2, -np.sqrt(3)/2, 5]], 12, 'rhombohedral'),
                ([[np.sqrt(3)/2, 1/2, 0], [0, 1, 0], [0, 0, 2]], 24, 'hexagonal'),
                ([[np.sqrt(3)/2, 1/2, 0], [0, 1, 0], [0, 0, 1]], 24, 'hexagonal equal length corner case'),
                ]
    for periods, n, name in lattices:
        group = bravais_point_group(periods, tr=False, ph=False)
        assert len(group) == n, (name, periods, group, n)
        group = bravais_point_group(periods @ R, tr=False, ph=False)
        assert len(group) == n, (name, periods, group, n)



import pytest
import warnings
import sympy
import itertools
import numpy as np
import scipy.linalg as la

from pytest import raises
from ..linalg import matrix_basis, sparse_basis
from ..groups import PointGroupElement


def test_sparse_basis():
    reals = [True, False]
    for _, real in itertools.product(range(3), reals):
        dim = np.random.randint(5, 10)
        H = np.random.rand(dim, dim)
        U = la.expm(1j*(H+H.T.conj()))
        # Sparsify a full rank matrix
        sparse_U = sparse_basis(U, num_digits=4, reals=real)
        assert sparse_U.shape == U.shape
        # Make one row linearly dependent on the others
        U[-1, :] = sum([np.random.rand()*row for row in U[:-1, :]])
        with warnings.catch_warnings(record=True) as w:
            sparse_U = sparse_basis(U, num_digits=4, reals=real)
            assert sparse_U.shape[0] == dim-1
            assert len(w) == 1 # A warning should be raised


def test_spatial_types():
    S1 = PointGroupElement(sympy.eye(2), False, False,
                           np.eye(3))
    S2 = PointGroupElement(sympy.Matrix([[0, 1], [1, 0]]), True, False,
                           np.eye(3))
    S3 = PointGroupElement(np.eye(2), False, False,
                           1j * np.eye(3))
    C6s = PointGroupElement(sympy.ImmutableMatrix(
                                [[sympy.Rational(1, 2), sympy.sqrt(3)/2],
                                 [-sympy.sqrt(3)/2,       sympy.Rational(1, 2)]]
                                                 ))
    C6f = PointGroupElement(np.array(
                                    [[1/2, np.sqrt(3)/2],
                                     [-np.sqrt(3)/2, 1/2]]
                                                     ))

    assert S2**2 == S1
    assert not S1 == S2
    assert S1 == S3
    assert C6s == C6f
    # Mixing sympy with other types raises an error
    with raises(ValueError):
        S = C6s * C6f
